/******************************************************************************
 * file:        SunADC.cc  
 * info:        read audio-device and transfer data to CSigInput class in the
 *              acoustic preprocessing CAkuPre via pipe
 * written by:  M.Beham, 12.01.95 (org. vor_onl.c by F.Flemisch)
 * last change: 
 * 
 ******************************************************************************/
/******************************************************************************
 *
 * Copyright (C) 1995 M. Beham
 *                                                                   
 * Alle Rechte vorbehalten. Diese Programmlistings und das daraus
 * erhaltene Programm duerfen nur mit schriftlicher Genehmigung	
 * des Autors weitergegeben werden.
 * Insbesondere duerfen auch Auszuege und die daraus erhaltenen
 * Programmteile nur mit schriftlicher Genehmigung des Autors
 * weitergeben, geaendert oder in Programmen anderer Autoren
 * verwendet werden.
 * Diese Copyright-Information darf nicht entfernt oder geaendert
 * werden.
 * 
 ******************************************************************************/
/* 
  Usage:

  SunADC <TransferBlockSize> <SamplingFrq>

  This programm runs under control of a parent-process as a child-process. It
  initializes the ADC audio device on a SUN-Station, starts a sampling process
  on request of the parent process and transfers sample data in blocks of
  fixed size to the parant process via a pipe which can be accessed in the 
  parent and child process (in the child stdout). 
  The parent-process realizes the CSigInputSnuADC class which is derived from
  CSigInput. Following signals are used for communication between the two
  processes:

  SIGUSR1   parent -> child    request of next sample-block 
  SIGUSR2   parent -> child    stop adc but keep data in buffer for read
  SIGINT    parent -> child    stop adc and clear buffer

  SIGUSR1   child  -> parent   adc is initialized and ready to start
  SIGUSR2   child  -> parent   buffer is empty, complete readout
*/

#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>
#include <stropts.h>
#include <fcntl.h> 
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <signal.h>
#include <poll.h>
#include <task.h>

#include <sys/audioio.h>

#define  MIN(a,b)    ( (a)<(b) ? a : b )
#define  PRINT_INFO
#define  FALSE               0
#define  TRUE                1


#define  ADC_DEVICE          "/dev/audio"
#define  ADC_PRECISION       16
#define  ADC_CHANNELS        1
#define  ADC_ENCODING        AUDIO_ENCODING_LINEAR
#define  ADC_FIFOLEN         5    // in seconds
#define  ALARM_TIMEOUT       5    // in seconds

// prg. name
static   char *prog = NULL; 



//-------------------------- static class CAudioDev ---------------------------

enum EState { Closed, Opened, Started, Stopped };

// init, start and stop of audio-device as stream
class CAudioDev
{
private:
    static char  *audiodev;
    static int    FileDes;
    static EState State;
public:
    CAudioDev( unsigned int samplefrq );
    ~CAudioDev();
    static int    getFD() { return FileDes; }
    static EState getState() { return State; }
    static void   start();
    static void   stop();
    static void   flush();
    static long   read( short *buf, long size );
};

// static initialization
int    CAudioDev::FileDes  = -1;
EState CAudioDev::State    = Closed;
char  *CAudioDev::audiodev = ADC_DEVICE;

// constructor: open and init audio-device
CAudioDev::CAudioDev( unsigned int samplefrq )
{
#ifdef PRINT_INFO
fprintf(stderr, "> %s: 'CAudioDev' constructor called\n", prog );
#endif
    // check if already opened
    if( State != Closed )
    {   fprintf( stderr, "FATAL %s: device %s already open\n", prog, audiodev);
        exit(1);
    }
   
    // get info to audio device
    struct stat  statinfo;
    int err = stat( audiodev, &statinfo );
    if( err < 0 )
    {   fprintf( stderr, "FATAL %s: can't get info (stat) to audio device: ", prog);
        perror(audiodev);
        exit(1);
    }
    if( !S_ISCHR(statinfo.st_mode) )
    {   fprintf( stderr, "FATAL %s: is not an audio device: %s\n", prog, audiodev );
        exit(1);
    }

    // open the audio device
    FileDes = open( audiodev, O_RDONLY | O_NDELAY );
    if( FileDes < 0 )
    {   switch(errno)
        {   case EBUSY:
            fprintf( stderr, "FATAL %s: device %s is busy\n", prog, audiodev );
            break;
            default:
            fprintf( stderr, "FATAL %s: can't open device: ", prog );
            perror(audiodev);
        }
        exit(1);
    }
    
    // try to get device info
    audio_device_t  devinfo;
    err = ioctl( FileDes, AUDIO_GETDEV, &devinfo );
    if( err < 0 )
    {   fprintf( stderr, "FATAL %s: can't control (ioctl) audio device: ", prog);
        perror(audiodev);
        exit(1);
    }
#ifdef PRINT_INFO
    fprintf(stderr, "> %s: device %s V(%s), cfg: %s\n", prog, devinfo.name, devinfo.version, devinfo.config );
#endif

    // try to get audio info
    audio_info_t adcinfo;
    err = ioctl( FileDes, AUDIO_GETINFO, &adcinfo );
    if( err < 0 )
    {   fprintf( stderr, "FATAL %s: can't get audio-info from device: ", prog);
        perror(audiodev);
        exit(1);
    }
    // try to set audio info and pause record
    adcinfo.record.sample_rate = adcinfo.play.sample_rate = samplefrq;
    adcinfo.record.channels    = adcinfo.play.channels  = ADC_CHANNELS;
    adcinfo.record.precision   = adcinfo.play.precision = ADC_PRECISION;
    adcinfo.record.encoding    = adcinfo.play.encoding  = ADC_ENCODING;
    adcinfo.record.pause       = 1;
    err = ioctl( FileDes, AUDIO_SETINFO, &adcinfo );
    if( err < 0 )
    {   fprintf( stderr, "FATAL %s: can't set audio-info to device: ", prog);
        perror(audiodev);
        exit(1);
    }
 
    // get audio info and compare with wanted parameter
    err = ioctl( FileDes, AUDIO_GETINFO, &adcinfo );
    if( err < 0 )
    {   fprintf( stderr, "FATAL %s: can't get audio-info from device: ", prog);
        perror(audiodev);
        exit(1);
    }
    if( adcinfo.record.sample_rate != samplefrq ||
        adcinfo.record.channels    != ADC_CHANNELS  ||
        adcinfo.record.precision   != ADC_PRECISION ||
        adcinfo.record.encoding    != ADC_ENCODING  ||
        adcinfo.record.pause       != 1 )
    {   fprintf( stderr, "FATAL %s: can't set parameter in audio-device: %s\n", prog, audiodev);
         exit(1);
    }
#ifdef PRINT_INFO
fprintf( stderr, "> %s: record-SampleRate %d, Channels %d, Precision %d, Encoding %d, Port %d\n", 
prog, adcinfo.record.sample_rate, adcinfo.record.channels, adcinfo.record.precision, adcinfo.record.encoding, adcinfo.record.port ); 
#endif

    // flush buffer and keep pause state ready for sampling
    flush();
}

// destructor: close device stream
CAudioDev::~CAudioDev()
{
#ifdef PRINT_INFO
fprintf(stderr, "> %s: 'CAudioDev' destructor called\n", prog );
#endif

    close(FileDes);
}

// start adc: resume from pause
void CAudioDev::start()
{
#ifdef PRINT_INFO
fprintf(stderr, "> %s: 'CAudioDev' start called\n", prog );
#endif

    if( State == Closed )
    {   fprintf( stderr, "FATAL %s: can't start ADC; device %s is not open\n", prog, audiodev );
        exit(1);
    }
    if( State == Started )
        return;

    // try to get audio info
    audio_info_t adcinfo;
    int err = ioctl( FileDes, AUDIO_GETINFO, &adcinfo );
    if( err < 0 )
    {   fprintf( stderr, "FATAL %s: can't get audio-info from device: ", prog);
        perror(audiodev);
        exit(1);
    }
    // reset pause flag
    adcinfo.record.pause = 0;
    err = ioctl( FileDes, AUDIO_SETINFO, &adcinfo );
    if( err < 0 )
    {   fprintf( stderr, "FATAL %s: can't set audio-info to device: ", prog);
        perror(audiodev);
        exit(1);
    }
    State = Started;
}

// stop audio dev
void CAudioDev::stop()
{
#ifdef PRINT_INFO
fprintf(stderr, "> %s: 'CAudioDev' stop called\n", prog );
#endif

    if( State == Closed || State == Opened )
    {   fprintf( stderr, "FATAL %s: can't stop ADC; device %s is not started\n", prog, audiodev );
        exit(1);
    }
    if( State == Stopped )
        return;

    // try to get audio info
    audio_info_t adcinfo;
    int err = ioctl( FileDes, AUDIO_GETINFO, &adcinfo );
    if( err < 0 )
    {   fprintf( stderr, "FATAL %s: can't get audio-info from device: ", prog);
        perror(audiodev);
        exit(1);
    }
    // reset pause flag
    adcinfo.record.pause       = 1;
    err = ioctl( FileDes, AUDIO_SETINFO, &adcinfo );
    if( err < 0 )
    {   fprintf( stderr, "FATAL %s: can't set audio-info to device: ", prog);
        perror(audiodev);
        exit(1);
    }
    State = Stopped;
}

// flush device buffer
void CAudioDev::flush()
{
    // flush buffer 
    int err = ioctl( FileDes, I_FLUSH, FLUSHR );
    if( err < 0 )
    {   fprintf( stderr, "FATAL %s: can't flush audio device: ", prog);
        perror(audiodev);
        exit(1);
    }
    State = Opened; 
}

// read audio device and check error
long CAudioDev::read( short *buf, long size )
{
    if( State != Started )
    {   fprintf( stderr, "FATAL %s: can't read audio-device; is not started\n", prog );
        exit(1);
    }

    // try to read maximum size and poll if not avail
    struct pollfd  PollSt;
    PollSt.fd = FileDes;
    PollSt.events = POLLIN;
    long rd = ::read(FileDes, buf, size*sizeof(short) );
    if( rd < 0 )
    {   if( State != Started ) return 0;
        switch(errno) 
        { case EAGAIN: poll(&PollSt, 1L, -1); return 0; break;
              default: fprintf(stderr, "FATAL %s: error reading audio device: ", prog);
                       perror(audiodev); exit(1);
        }
    }
 
    // check record status
    audio_info_t adcinfo;
    int err = ioctl( FileDes, AUDIO_GETINFO, &adcinfo );
    if( err < 0 )
    {   fprintf( stderr, "FATAL %s: can't get audio-info", prog);
        perror("; ");
        exit(1);
    }
    if( adcinfo.record.error != 0 )
    {   fprintf( stderr, "FATAL %s: audio device record error\n", prog);
        exit(1);
    }

    long cnt = rd / sizeof(short);
    // check even count
    if( cnt * sizeof(short) != rd )
    {   fprintf( stderr, "FATAL %s: audio device odd number of samples\n", prog);
        exit(1);
    }

    return cnt;
}

//-------------------------- classes for Signal Handler ---------------------


// interupt handler for all signals:

// signal SIGUSR1: block request
class CSigRequest : public Interrupt_handler
{
private:
    int    RequestCount;
    void   interrupt( ) { RequestCount++; }
public:
    CSigRequest( ) : Interrupt_handler(SIGUSR1), RequestCount(0) { }
    int    getCount() const { return RequestCount; }
    void   decCount() { RequestCount--; }
    void   resetCount() { RequestCount = 0; }
};

// signal SIGUSR2: stop adc
class CSigStopADC : public Interrupt_handler
{
private:
    int    ADCPause;
    void   interrupt() { ADCPause = TRUE; CAudioDev::stop(); }
public:
    CSigStopADC( ) : Interrupt_handler(SIGUSR2), ADCPause(FALSE) { }
    int    getPause() const { return ADCPause; }
    void   resetPause() { ADCPause = FALSE; }
};

// signal SIGINT: stop adc and clear buffer
class CSigClearADC : public Interrupt_handler
{
private:
    int    ADCClear;
    void   interrupt() { ADCClear = TRUE; CAudioDev::stop(); }
public:
    CSigClearADC( ) : Interrupt_handler(SIGINT), ADCClear(FALSE) { }
    int    getClear() const { return ADCClear; }
    void   resetClear() { ADCClear = FALSE; }
};

// signal SIGALRM: timeout; test if parent is alive
class CSigTimeout : public Interrupt_handler
{
private:
    void   interrupt(); 
public:
    CSigTimeout( ) : Interrupt_handler(SIGALRM) { }
};

void CSigTimeout::interrupt()
{ 
    if( getppid() == 1 )
    {   fprintf(stderr, "> %s: exit, because parent-process exited\n", prog );
        exit(0);
    } 
    alarm( ALARM_TIMEOUT ); 
}

//----------------------------- class CSigFifo --------------------------------

class CSigFifo
{
private:
    const  long   FifoSize;
    const  long   BlockSize;
    short * const Head;
    short * const Tail;
    short  *Wr;
    short  *Rd;
    long    NSamples;
    int     Overflow;
public:
    CSigFifo( long blocksize, long samplefrq );
    ~CSigFifo();
    int writeSamples( );
    int readBlock( int toFD );
    int readDummy( int fd );
    int isOverflow() const { return Overflow; } 
    void reset();
};

// Fifo constructor
CSigFifo::CSigFifo( long blocksize, long samplefrq )
:   BlockSize(blocksize),
    FifoSize((long)( (samplefrq*ADC_FIFOLEN) / blocksize ) * blocksize),
    Head(new short[FifoSize]),
    Tail(Head + FifoSize)
{
    if( Head == 0 )
    {   fprintf( stderr, "FATAL %s: out of memory\n", prog);
        exit(1);
    }
    reset();
}

// Fifo destructor
CSigFifo::~CSigFifo( )
{
    delete [] Head;
}

// reset Fifo
void CSigFifo::reset()
{
    Wr = Rd = Head;
    NSamples  = 0;
    Overflow = FALSE;
}   

// write maximum of avail. samples from stream to fifo
int CSigFifo::writeSamples( )
{
    // check fifo overflow
    long rest = FifoSize - NSamples;
    if( rest <= 0 )
    {   Overflow = TRUE;
        CAudioDev::stop();
        fprintf(stderr, "WARNING %s: signal fifo overflow\n", prog );
        return(0);
    }

    long cnt = CAudioDev::read( Wr,  MIN(rest, Tail-Wr) );
  
    // now max samples from queue is read
    NSamples += cnt;
    Wr       += cnt;
    if( Wr >= Tail ) Wr = Head;
    return 1;
}

// read one block of samples from fifo to stream
int  CSigFifo::readBlock( int fd )
{
    // check if one block is in Fifo
    if( NSamples < BlockSize || Tail - Rd < BlockSize)
        return 0;

    // copy one block from Fifo to stream
    long cnt = write( fd, Rd, BlockSize*sizeof(short) );
    if( cnt != BlockSize*sizeof(short) )
    {    fprintf(stderr, "FATAL %s: can't write block of samples to stream", prog );
         perror(", "); exit(1);
    }
    NSamples -= BlockSize;
    Rd += BlockSize;
    if( Rd >= Tail ) Rd = Head;
    return BlockSize;
}

// read dummy block
int  CSigFifo::readDummy( int fd )
{
    // copy one block from Fifo to stream
    long cnt = write( fd, Head, BlockSize*sizeof(short) );
    if( cnt != BlockSize*sizeof(short) )
    {    fprintf(stderr, "FATAL %s: can't write samples to stream", prog );
         perror(", "); exit(1);
    }
    return BlockSize;
}


//--------------------------------- main --------------------------------------

main( int argc, char *argv[] )
{
    prog = argv[0];
    int  SendSignals = TRUE;
    int  outFD = 1;  // stdout

    // check arguments
    if( argc < 3 || argc > 4)
    {   fprintf( stderr, "usage: %s <BlockSize> <SampleFrq> [s]\n", prog );
        fprintf( stderr, " s: (optional) no signals to parent process\n" );
        exit(1);
    }
    long adc_blocksize = atol( argv[1] );
    long adc_samplefrq = atol( argv[2] );
    if( argc == 4 && *argv[3] == 's' ) SendSignals = FALSE;
#ifdef PRINT_INFO
fprintf(stderr, "> %s: ProcessId %d, BlockSize %d\n", prog, getpid(), adc_blocksize );
#endif  

    // init fifo signal buffer 
    CSigFifo  Fifo( adc_blocksize, adc_samplefrq );
 
    // proc. id of parent process
    int ParentPid = getppid();

    // init handler for signals
    CSigRequest  SigRequest;
    CSigStopADC  SigStopADC;
    CSigClearADC SigClearADC;
    CSigTimeout  SigTimeout;
    // start timeout watchdog
    alarm(ALARM_TIMEOUT);

    // open audio device
    CAudioDev AudioDev(adc_samplefrq);

    // endless loop for start and stop and read
    for( ; ; )
    {
        // reset Fifo and Signal-Handler
        Fifo.reset();
        SigRequest.resetCount();
        SigStopADC.resetPause();
        SigClearADC.resetClear();

        // now child is ready for first request
        if( SendSignals ) kill( ParentPid, SIGUSR1 );

        // wait for first request
        while( SigRequest.getCount() == 0 )
            pause();

        // now start ADC
	AudioDev.flush() ;
        AudioDev.start();

        // recording loop until end
        do
        {   // check buffer overflow and signal stop ADC
            if( !Fifo.isOverflow()  && AudioDev.getState() == Started )
            {   // copy samples to Fifo
                Fifo.writeSamples( );
            } else
            {   // wait for next request
                while( SigRequest.getCount() == 0 && !SigClearADC.getClear() )
                    pause();
            }
            // copy block to output until end of buffer
            if( SigRequest.getCount() > 0 )
            {   if( Fifo.readBlock( outFD ) > 0 )
                    SigRequest.decCount();
                else if( AudioDev.getState() == Stopped )
                    break;
            }
        } while( !SigClearADC.getClear() );

        // check end condition
        if( !SigClearADC.getClear() )
            if( SendSignals ) kill( ParentPid, SIGUSR2 );
        // if request, copy dummy to output stream
        if( SigRequest.getCount() > 0 )
            Fifo.readDummy( outFD );
     } // for ever loop

    return 0;
}

