/******************************************************************************
 * file:        akupre.cc  
 * info:        implementation of CAkuPre (Acoustic Preprocessing)  
 * written by:  M.Beham, 12.01.95 (new)
 * last change: 
 * 
 ******************************************************************************/
/******************************************************************************
 *
 * Copyright (C) 1995 M. Beham
 *                                                                   
 * Alle Rechte vorbehalten. Diese Programmlistings und das daraus
 * erhaltene Programm duerfen nur mit schriftlicher Genehmigung	
 * des Autors weitergegeben werden.
 * Insbesondere duerfen auch Auszuege und die daraus erhaltenen
 * Programmteile nur mit schriftlicher Genehmigung des Autors
 * weitergeben, geaendert oder in Programmen anderer Autoren
 * verwendet werden.
 * Diese Copyright-Information darf nicht entfernt oder geaendert
 * werden.
 * 
 ******************************************************************************/
/* 
Beschreibung:    
  
*/

#include <string.h>
#include <math.h>

// #define DEBUG_LEVEL (3)
// #define AP_TEST 

#include "dbg_mess.h"

// include right akup_xxx.h
#if       defined(FEAT_MUC)
#include "akup_muc.h"
#elif     defined(FEAT_DRS)
#include "akup_drs.h"
#else
#error    no match for akup_xxx.h include
#endif



//--------------------------- public 'CAkuPre' -----------------------------


// constructor: CAkuPre
CAkuPre::CAkuPre( FILE *channelfile )
 :  ReadMode(  "AkuPre.ReadMode", ESPEECHFLAG, channelfile, U ),
    DeviceStr( "AkuPre.Device", channelfile, "???" ),
    SampleShift( "AkuPre.SampleShift", channelfile, 0 ),
    Preemphase( "AkuPre.Preemphase", channelfile, 0.0 ),
#ifdef FEAT_DRS
    NoisePower( "AkuPre.NoiseEngDB", channelfile, 0.0 ),
    QuantPower( "AkuPre.QuantEngDB", channelfile, 0.0 ),
#endif
    SigAnl( WINDOW_DELTA, WINDOW_LENGTH, FFT_LENGTH, DeviceStr.Par, SampleShift.Par, Preemphase.Par ),
#ifdef FEAT_DRS
    PreFeature( FFT_LENGTH/2, (double)SAMPLE_FRQ, NoisePower.Par, QuantPower.Par ),
#endif
#ifdef FEAT_MUC
    PreFeature( FFT_LENGTH/2, (double)SAMPLE_FRQ ),
#endif
    SpNonSpOverlap( "AkuPre.UseSpNonSpOverlap", channelfile, 0 ),
    //SpNonSp( PreFeature.getSpDetFeatureLen(), (double)WINDOW_DELTA/(double)SAMPLE_FRQ, channelfile ),
    SpNonSp(0),
    GainControl( channelfile, (double)WINDOW_DELTA/(double)SAMPLE_FRQ ),
    ChanComp( PRESPECLEN, channelfile, (double)WINDOW_DELTA/(double)SAMPLE_FRQ ),
    Delta( PRESPECLEN, DELTASPECLEN ),
    BufLen( FEATURE_BUFLEN ),
    DeltaF( DELTA_FORW ),
    DeltaB( DELTA_BACK ),
    PastHold( 2*DeltaB + 1 ),
    MinBufFill( 2*DeltaF + 2 ), // kor. 11.08.95 2*deltaF+1 M.Beh
    FrqSamp( SAMPLE_FRQ )
{
    int i;
    DEBUG_1_MSG("'CAkuPre' constructor called" );
    fprintf( stderr, "FFT-Len(komplex) %d\n", FFT_LENGTH ) ;
    fprintf( stderr, "Window-Length    %d\n", WINDOW_LENGTH ) ;
    fprintf( stderr, "Window-Delta     %d\n", WINDOW_DELTA ) ;
    fprintf( stderr, "Sample-Freq      %d\n", SAMPLE_FRQ ) ;

    IF_FATAL_EXIT( PastHold+MinBufFill >= BufLen , "'CAkuPre' past hold + min. buffer fill > buffer length" );

    // SpNonSp
    if( !SpNonSpOverlap.Par )  
      SpNonSp = new CSpeechNonSpeech( PreFeature.getSpDetFeatureLen(), (double)WINDOW_DELTA/(double)SAMPLE_FRQ, channelfile );
    else
      SpNonSp = new CSpeechNonSpeechOverlap( PreFeature.getSpDetFeatureLen(), (double)WINDOW_DELTA/(double)SAMPLE_FRQ, channelfile );
	
    // alloc buffer
    BasePtr = new SFeatBufEntry[BufLen];
    IF_FATAL_EXIT( BasePtr == 0, "'CAkuPre' out of memory" );

    // set norm-factor for hamming window
    // double shiftFactor = ldexp( 1.0, SampleShift.Par );
    SigAnl.normWindow( PreFeature.getNormFactor() /* / shiftFactor */ );

    // next ptr
    for( i=0; i<BufLen-1; i++ )
        BasePtr[i].Next = &BasePtr[i+1];
    BasePtr[BufLen-1].Next = &BasePtr[0];
    // last ptr
    BasePtr[0].Last = &BasePtr[BufLen-1];
    for( i=1; i<BufLen; i++ )
        BasePtr[i].Last = &BasePtr[i-1];
    // delta n ptr
    for( i=0; i<BufLen; i++ )
    {   int dind;
        // forward
        dind = i + DeltaF;
        if( dind >= BufLen )   dind -= BufLen;
        else if( dind < 0 )    dind += BufLen;
        BasePtr[i].DeltaForw = &BasePtr[dind];
        // backward
        dind = i - DeltaB;
        if( dind >= BufLen )   dind -= BufLen;
        else if( dind < 0 )    dind += BufLen;
        BasePtr[i].DeltaBack = &BasePtr[dind];
    }
    resetBuf();
}

// destructor: CAkuPre
CAkuPre::~CAkuPre( )
{
    DEBUG_1_MSG("'CAkuPre' destructor called" );
    
    if( SpNonSp != 0 )
      delete SpNonSp ;

    // delete buffers
    delete BasePtr;
}

// start AkuPre
int  CAkuPre::startAkuPre( const char *filename )
{
    DEBUG_2_MSG_S("'CAkuPre' startAkuPre called with:", filename );

    // clear PreFeature buffer
    resetBuf();
    // start signal analyse 
    IF_ERROR_RETURN( SigAnl.startAnl( filename ) != _OK_, "'CAkuPre' startAkuPre failed" );
    // check sample-frq.
    IF_ERROR_RETURN_S( SigAnl.getFrqSamp() != FrqSamp, "'CAkuPre' startAkuPre: sampling frq. of", filename)

    // init speech non speech detector (alpha recursion)
    SpNonSp->start();

    // init buffer to min. fill
    WrState = Started;
    WPtr->Last->Num = -1;
    while( BufFill < MinBufFill-1 ) 
    {   // next power spec. calc if possible 
        if( writeEntry() == 0 )
        {   ERROR_MSG_S("'CAkuPre' startAkuPre: can't fill delta-buffer from:", filename ); 
            return _ERR_;
        }
    }
    // copy to past
    SFeatBufEntry *lp = RPtr;
    for( int i=0; i<PastHold-1; i++ )
    {   lp = lp->Last;
        memcpy( &lp->Vec.Pre, &RPtr->Vec.Pre, sizeof(SPreFeature) );
        lp->Num = -i-1;
    }
    RdState = Started;
    WrState = Running;
    return _OK_;
}

// stop AkuPre
void CAkuPre::stopAkuPre()
{
    DEBUG_2_MSG("'CAkuPre' stopAkuPre called" );

    // stop signal analyse
    SigAnl.stopAnl();
}

// clear AkuPre
void CAkuPre::clearAkuPre()
{
    DEBUG_2_MSG("'CAkuPre' clearAkuPre called" );

    // stop signal analyse
    SigAnl.clearAnl();
    // clear Feature buffer
    resetBuf();
}

// get next pre. proc. feature vector
SPreFeature *CAkuPre::getNextPre()
{
    DEBUG_4_MSG("'CAkuPre' getNextPre called" );

    // switch state
    if( RdState == Running )
    {   ERROR_MSG( "'CAkuPre' getNextPre: post proc. already running" );
        return 0;
    }
    if( RdState == Stopped )
    {   ERROR_MSG( "'CAkuPre' getNextPre: AkuPre is not started" );
        return 0;
    }

    RdState = Preproc;

    // write one spec to buffer
    if( BufFill < 1 && writeEntry() == 0 )
        return 0;
    // read out this spec
    if( BufFill-- < 0 )
    {   ERROR_MSG("'CAkuPre' getNextPre: buffer underflow"); 
        BufFill++;
        WPtr = WPtr->Next;       
    }
    RdTotal++;
    RPtr = RPtr->Next;
#if defined(DEBUG_LEVEL) && (DEBUG_LEVEL>=3 || DEBUG_LEVEL==-3)
fprintf( stderr, ">  readEntry: Num=%2d, BufFill=%2d,  WrTotal=%2d, RdTotal=%2d, UndefCount=%2d, Flag=%1d, RdState=%1d\n", 
          RPtr->Last->Num, BufFill, WrTotal, RdTotal, UndefCount, RPtr->Last->Flag, RdState);
#endif    
    return &RPtr->Last->Vec.Pre;
}


// get next vector
SFeature *CAkuPre::getNext()
{
    DEBUG_4_MSG("'CAkuPre' getNext called" );
    int            i;
    SFeatBufEntry *ep;

    // switch state
    switch( RdState )
    {
    case Started:
        // calc. new spec if not min. buffer filled or wanted is undef.
        while( BufFill < MinBufFill || getRd() != Wanted ) // kor. <= ins. <
	    {   if( (BufFill < MinBufFill || getRd() == Undef) && writeEntry() == 0 ) // kor. <= ins. <
            {   RdState = Stopped;
                return 0;
            } 
            if( getRd() == NotWanted )
            {   // read out not wanted spec
                readDummy();
            }
        }
        // spec wanted: post pocessing 
        for( ep=RPtr, i=0; i<2*DeltaB; i++ )
        {   ep = ep->Last;
            ChanComp.subMean( ep->Vec.Pre.Spec, ep->Flag );  
            PreFeature.postPreSpec( &ep->Vec.Pre );
            GainControl.norm( &ep->Vec.Pre, &ep->Vec.Norm, ep->Flag );
        }
        for( ep=RPtr, i=0; i<2*DeltaF; i++ )
	    {   //SpNonSp.update( PreFeature.getSpDetFeature(&ep->Vec.Pre), ep->Flag );
	    SpNonSp->update( PreFeature.getSpDetFeature(&ep->Vec.Pre), ep->Flag, ep->NrUndef, ep->InUndef);

            ChanComp.subMean( ep->Vec.Pre.Spec, ep->Flag ); 
            PreFeature.postPreSpec( &ep->Vec.Pre );
            GainControl.norm( &ep->Vec.Pre, &ep->Vec.Norm, ep->Flag );
            ep = ep->Next;
        }
        // spectrum wanted: calc deltas 
        for( ep=RPtr, i=0; i<DeltaB; i++ )
        {   ep = ep->Last;
            Delta.calcDelta( ep );
        }
        for( ep=RPtr, i=0; i<DeltaF; i++ )
        {   Delta.calcDelta( ep );
            ep = ep->Next;
        }
        RdState = Running;
        return readEntry();
        break;
    case Running:
        // calc. new spec if not min. buffer filled or wanted is undef.
        if( BufFill < MinBufFill && writeEntry() == 0 ) 
        {   RdState = Stopped;
            return 0;
        }            
        while( getRd() == Undef )
        {   IF_FATAL_EXIT( writeEntry() == 0, "'CAkuPre' getNext: write entry is stopped incorrect");
        }
        // now wanted is known,  read out
        if( getRd() == Wanted )
        {   return readEntry();
        } else
        {   RdState = Started;
            readDummy();
            return 0;
        }
        break;
    default:
        ERROR_MSG("'CAkuPre' getNext: read AkuPre is stopped");
        return 0;
    } // switch RdState

}


//////////////////////////////////////////////////////////////////////////////
//
// appending new cfg-file at the end of the old one!
//       old cfg-file given by filepointer fp
//
/////////////////////////////////////////////////////////////////////////////
int CAkuPre::appendPar( FILE *fp ) {
    int ret = 0 ;
   
    // append fix parameters
    ret |= appendAkuPrePar( fp ) ;
    fprintf( fp, "\n" ) ;
    ret |= SpNonSp->appendFixPar( fp ) ;
    fprintf( fp, "\n" ) ;
    ret |= ChanComp.appendFixPar( fp ) ;
    fprintf( fp, "\n" ) ;
    ret |= GainControl.appendFixPar( fp ) ;
    fprintf( fp, "\n" ) ;
    // append updatable params
    ret |= SpNonSp->appendUpdPar( fp ) ;
    fprintf( fp, "\n" ) ;
    ret |= ChanComp.appendUpdPar( fp ) ;
    fprintf( fp, "\n" ) ;
    ret |= GainControl.appendUpdPar( fp ) ;

    fprintf( fp, "\n" ) ;

    return( ret == 0 ? ret : -1 ) ;
   
} ; // END appendPar
    
    



//--------------------------- private 'CAkuPre' -------------------------------


// write vector and act. ptr
SFeatBufEntry *CAkuPre::writeEntry( )
{
    DEBUG_4_MSG("'CAkuPre' writeEntry called" );

    // check if stopped
    if( WrState == Stopped )
        return 0;
    // signal analysis: get powerspec and zcr
    float  zcrate;
    float *powspec = SigAnl.getNextSpec( &zcrate );
    if( powspec != 0 )
    {   // pre processing of powerspec 
        PreFeature.convPowSpec( powspec, zcrate, &WPtr->Vec.Pre );
        WPtr->Flag = SpNonSp->detect( PreFeature.getSpDetFeature(&WPtr->Vec.Pre) ); 
	WPtr->ProbSp = SpNonSp->getProbSp() ;
	WPtr->ProbSi = SpNonSp->getProbSi() ;
    } else {
        WPtr->Flag = P;
	WPtr->ProbSp = -1.0 ;
	WPtr->ProbSi = -1.0 ;
    }    

    // act. speech flags in the past (only here)
    ESpeechFlag flag ;
    if( WPtr->Flag != U ) {   
      if( SpNonSpOverlap.Par ) {
	if( LastFlag == P ) 
	  flag = ( WPtr->Flag==P ? PUP : ( WPtr->Flag==S ? PUS : PUO ) ) ;
	else if( LastFlag == S ) 
	  flag = ( WPtr->Flag==P ? SUP : ( WPtr->Flag==S ? SUS : SUO ) ) ;
	else if( LastFlag == O ) 
	  flag = ( WPtr->Flag==P ? OUP : ( WPtr->Flag==S ? OUS : OUO ) ) ;
      } else 
	flag = LastFlag==P ? (WPtr->Flag==P ? PUP : PUS) : (WPtr->Flag==P ? SUP : SUS);
      
      SFeatBufEntry *pp = WPtr;
      for( int i=0; i < UndefCount; i++ )
        {   pp = pp->Last;
	pp->Flag = flag;
	pp->NrUndef = UndefCount;
	pp->InUndef = UndefCount-i-1;
        }
      UndefCount = 0;
      LastFlag = WPtr->Flag;
    } else { 
      UndefCount++;
    }
    
    // update WPtr
    if( powspec != 0 ) {
      WPtr->Num = WPtr->Last->Num+1;
      WPtr = WPtr->Next;
      BufFill++;
      WrTotal++;
    } else {   // copy dummy to future
      SFeatBufEntry *lp = WPtr->Last;
      for( int i=0; i<MinBufFill-1; i++ ) {
	memcpy( &WPtr->Vec.Pre, &lp->Vec.Pre, sizeof(SPreFeature) );
	WPtr->Num = WPtr->Last->Num+1;
	WPtr = WPtr->Next;
	BufFill++;
      }
      WrState = Stopped;    
    }
    
    if( BufFill + PastHold >= BufLen ) {
      ERROR_MSG("'CAkuPre' writeEntry: buffer overflow"); 
      BufFill--;
      RPtr = RPtr->Next;
    }

#if defined(DEBUG_LEVEL) && (DEBUG_LEVEL>=2 || DEBUG_LEVEL==-2)
    fprintf( stderr, "> writeEntry: Num=%2d, BufFill=%2d,  WrTotal=%2d, RdTotal=%2d, UndefCount=%2d, Flag=%1d, WrState=%1d\n", 
	     WPtr->Last->Num, BufFill, WrTotal, RdTotal, UndefCount, WPtr->Last->Flag, WrState);
#endif
    return WPtr->Last;
}

// post proc., read vector and act. ptr
SFeature *CAkuPre::readEntry( )
{
    DEBUG_4_MSG("'CAkuPre' readEntry called" );
    if( BufFill-- < MinBufFill ) 
    {   ERROR_MSG("'CAkuPre' readEntry: buffer underflow"); 
        BufFill++;
        WPtr = WPtr->Next;       
    }
    // post processing
    SFeatBufEntry *fowRP = RPtr->DeltaForw->DeltaForw;
    SpNonSp->update( PreFeature.getSpDetFeature(&fowRP->Vec.Pre), fowRP->Flag, fowRP->NrUndef, fowRP->InUndef);
    ChanComp.subMean( fowRP->Vec.Pre.Spec, fowRP->Flag );
    PreFeature.postPreSpec( &fowRP->Vec.Pre );
    GainControl.norm( &fowRP->Vec.Pre, &fowRP->Vec.Norm, fowRP->Flag );
    Delta.calcDelta(  RPtr->DeltaForw );
    Delta.calcDDelta( RPtr );
    RdTotal++;
    RPtr = RPtr->Next;
#if defined(DEBUG_LEVEL) && (DEBUG_LEVEL>=3 || DEBUG_LEVEL==-3)
 fprintf( stderr, ">  readEntry: Num=%2d, BufFill=%2d,  WrTotal=%2d, RdTotal=%2d, UndefCount=%2d, Flag=%1d, RdState=%1d\n", 
          RPtr->Last->Num, BufFill, WrTotal, RdTotal, UndefCount, RPtr->Last->Flag, RdState);
#endif    
    return &RPtr->Last->Vec;
}

// dummy read vector and act. ptr
SFeature *CAkuPre::readDummy( )
{
    DEBUG_4_MSG("'CAkuPre' readEntry called" );
    if( BufFill-- < MinBufFill )
    {   ERROR_MSG("'CAkuPre' readEntry: buffer underflow"); 
        BufFill++;
        WPtr = WPtr->Next;       
    }
    // post proc. only update spnonsp-detector
    SFeatBufEntry *fowRP = RPtr->DeltaForw->DeltaForw;
    // SpNonSp.update( PreFeature.getSpDetFeature(&fowRP->Vec.Pre), fowRP->Flag );
    SpNonSp->update( PreFeature.getSpDetFeature(&fowRP->Vec.Pre), fowRP->Flag, fowRP->NrUndef, fowRP->InUndef);
    RPtr = RPtr->Next;
#if defined(DEBUG_LEVEL) && (DEBUG_LEVEL>=3 || DEBUG_LEVEL==-3)
 fprintf( stderr, ">  readDummy: Num=%2d, BufFill=%2d,  WrTotal=%2d, RdTotal=%2d, UndefCount=%2d, Flag=%1d, RdState=%1d\n", 
          RPtr->Last->Num, BufFill, WrTotal, RdTotal, UndefCount, RPtr->Last->Flag, RdState);
#endif    
    return &RPtr->Last->Vec;
}

// get read wanted info
ERdFlag  CAkuPre::getRd()
{
    return ReadMode.Par == U    ? Wanted : 
           ( RPtr->Flag == U    ? Undef  : 
           ( RPtr->Flag >= ReadMode.Par ? Wanted : NotWanted ) );
}

// reset and clear buffer
void  CAkuPre::resetBuf() 
{ 
    WPtr = RPtr = BasePtr; 
    UndefCount = BufFill = RdTotal = WrTotal = 0;
    LastFlag = P;
    RdState = WrState = Stopped;  
    for( int i=0; i<BufLen; i++)
        BasePtr[i].Flag  = P;
}


//////////////////////////////////////////////////////////////////////////////
//
// append params for AkuPre
//
/////////////////////////////////////////////////////////////////////////////
int CAkuPre::appendAkuPrePar( FILE *fp ) {

    int ret = 0 ;
    fprintf( fp, "// 1. AkuPre.  general parameters for preprocessing\n" ) ;
    ret |= DeviceStr.appendPar( fp ) ;
    ret |= ReadMode.appendPar( fp ) ;
    ret |= SampleShift.appendPar( fp ) ;
    ret |= Preemphase.appendPar( fp ) ;

    return( ret == 0 ? ret : -1 ) ;
} ; // END appendApMainPar



//--------------------------- end of 'CAkuPre' --------------------------------

//--------------------------- test --------------------------------------------

#ifdef AP_TEST

#include <stdlib.h>

int main( int argc, char *argv[] )
{   
    IF_ERROR_RETURN( argc != 3, "usage: <chanfile> <samplefile>" );  
    FILE *chanfp = fopen( argv[1], "r+" );
    IF_ERROR_RETURN_S( chanfp == 0, "can't open channel-file:", argv[1]);

    CAkuPre ap( chanfp );
    SFeature  *vf;

    if( ap.startAkuPre( argv[2] ) == _ERR_ ) return(-1);

    for( int i=0; ap.testRdIsRun(); i++) 
    {   if((vf = ap.getNext( )) != 0)
        {   printf( "%d. \n", ap.getRdNum() );
            ap.printFeature( vf );
        }
    }
    ap.clearAkuPre();

    return(0);
}

#endif  // TEST_LEVELcd source
