/******************************************************************************
 * file:        fft.cc  
 * info:        source for CFastFourier (FFT)
 * written by:  M.Beham, 12.01.95 (org. from fft.c)
 * last change: 
 * 
 ******************************************************************************/
/******************************************************************************
 *
 * Copyright (C) 1995 M. Beham
 *                                                                   
 * Alle Rechte vorbehalten. Diese Programmlistings und das daraus
 * erhaltene Programm duerfen nur mit schriftlicher Genehmigung	
 * des Autors weitergegeben werden.
 * Insbesondere duerfen auch Auszuege und die daraus erhaltenen
 * Programmteile nur mit schriftlicher Genehmigung des Autors
 * weitergeben, geaendert oder in Programmen anderer Autoren
 * verwendet werden.
 * Diese Copyright-Information darf nicht entfernt oder geaendert
 * werden.
 * 
 ******************************************************************************/
/* 
Beschreibung:    Fast-Fourier-Transformation (FFT)
Funktionen zur diskreten Fourietransformation und inversen diskreten Fourier-
transformation eines komplexen Vektors x(i) mit N Elementen (wobei N=2^b)
in den ebenfalls komplexen Vektor X(j) (N Elemente) nach dem Butterfly-
algorithmus (FFT = Fast-Fourier-Transformation). Die Transformation eines
Vektors mit N Elementen (N ist 2er Potenz) benoetigt N/2*ld(N) komplexe
Multiplikationen (= 2N*ld(N) reelle Mul.). Die komplexen Drehzeiger werden
einmalig mit dem Constructor der Klasse berechnet und in einer Tabelle 
abgelegt (sortiert in Bit-Reversal-Indexorder passend zum FFT-Algorithmus,
nur cos-Tabelle).  
*/


#include <math.h>

// #define DEBUG_LEVEL 4
// #define TEST_LEVEL 

#include "dbg_mess.h"
#include "fft.h"


////////// public


// constructor: CFastFourier
CFastFourier::CFastFourier( unsigned int winlen )
: Order( (int)logb( (double) winlen ) ), Len( winlen ) 
{
    DEBUG_1_MSG("'CFastFourier' constructor called" );
    IF_FATAL_EXIT( (int)(ldexp(1.0, Order)+0.5) != Len, "'CFastFourier' N is not power of 2" );
    IF_FATAL_EXIT( Order < 2 || Order > 16, "'CFastFourier'  N < 4 or N > 2^16" );

    // cos/sin table (butterfly ops) and bit-revers-tab calloc and calc
    BfRePtr  = new float[Len/2];
    BfImPtr  = new float[Len/2];
    SpBitPtr = new unsigned short[Len];
    IF_FATAL_EXIT( BfRePtr == 0 || BfImPtr == 0 || SpBitPtr == 0, "'CFastFourier' out of memory");

    unsigned short a;
    unsigned short p;
    for( p = 0; p < Len; p++ )
    {   SpBitPtr[p] = 0;
        a = p;
        for( unsigned int i = 0; i < Order; i++, a>>=1 )
        {   SpBitPtr[p] <<= 1; SpBitPtr[p] |= a & 01; }
    }

    double pi2n = -2.0 * M_PI / Len, phi;
    for( p = 0; p < Len/2; p++ )
    {   phi = pi2n * (double)SpBitPtr[p<<1];
        BfRePtr[p] = cos(phi);
        BfImPtr[p] = sin(phi);
    }

}

// destructor: CFastFourier
CFastFourier::~CFastFourier()
{
    DEBUG_1_MSG("'CFastFourier' destructor called" );

    delete [] BfRePtr;
    delete [] BfImPtr;
    delete [] SpBitPtr;
}

// Transformation
void CFastFourier::transform( float *RePtr, float *ImPtr )
{
    DEBUG_4_MSG("'CFastFourier' transform called" );

    float            TmpRe, TmpIm;	/* komplexer Zwischenspeicher */
    register float  *bfptrre, *bfptrim;	/* Butterflyoperator-Tab.Ptr (re:cos, im:sin) */
    register float  *p_re, *p_im;	/* Pointer in den Vektor (Abstand d) */
    register int     d;			/* Var. 1. Schleife: N/2, N/4, ..., 1  */
    register int     k;			/* Var. 2. Schleife: N/2, N/2-1, ..., 1 */
    register int     modulo_d;		/* Zaehler Modulo d (1,2...d,1,2..) */

    /* Butterflyalgorithmus: 
    * 1. Schleife ld(n) mal, d = n/2, n/4, ... 1
    * 2. Schleife n/2 mal, k_1 = 0, 1, ... d-1, 2d, 2d+1, ...
    * insgesammt n/2 ld(n) komplexe Multiplikationen
    */
    for( d = Len>>1; d ; d >>= 1 ) 
    {
    	for( k = Len>>1, modulo_d = d, p_re = RePtr, p_im = ImPtr, bfptrre = BfRePtr, bfptrim = BfImPtr;
    	     k ; k--, p_re++, p_im++ )
    	{
    	    TmpRe   = *bfptrre * p_re[d] - *bfptrim * p_im[d];
    	    TmpIm   = *bfptrre * p_im[d] + *bfptrim * p_re[d];
            p_re[d] = *p_re - TmpRe;
            p_im[d] = *p_im - TmpIm;
            *p_re   = *p_re + TmpRe;
            *p_im   = *p_im + TmpIm;

    	    if( !(--modulo_d) ) { p_re += d; p_im += d; bfptrre++; bfptrim++; modulo_d = d; }
      	}
    }
    
    /* Ergebnisvektor umsortieren in Bit-Reversal Order */
    for (d = 0; d < Len; d++)
    {   if ( (k = SpBitPtr[d]) > d )
      	{    TmpRe = RePtr[d];
             TmpIm = ImPtr[d];
             RePtr[d] = RePtr[k];
             ImPtr[d] = ImPtr[k];
             RePtr[k] = TmpRe;
             ImPtr[k] = TmpIm;
      	}
    }

}

// invers transformation ifft
void CFastFourier::invers( float *RePtr, float *ImPtr )
{
    DEBUG_4_MSG("'CFastFourier' invers called" );

    float            TmpRe, TmpIm;	/* komplexer Zwischenspeicher */
    register float  *bfptrre, *bfptrim;	/* Butterflyoperator-Tab.Ptr (re:cos, im:sin) */
    register float  *p_re, *p_im;	/* Pointer in den Vektor (Abstand d) */
    register int     d;			/* Var. 1. Schleife: N/2, N/4, ..., 1  */
    register int     k;			/* Var. 2. Schleife: N/2, N/2-1, ..., 1 */
    register int     modulo_d;		/* Zaehler Modulo d (1,2...d,1,2..) */

    /* Butterflyalgorithmus: 
    * 1. Schleife ld(n) mal, d = n/2, n/4, ... 1
    * 2. Schleife n/2 mal, k_1 = 0, 1, ... d-1, 2d, 2d+1, ...
    * insgesammt n/2 ld(n) komplexe Multiplikationen
    */
    for( d = Len>>1; d ; d >>= 1 ) 
    {
    	for( k = Len>>1, modulo_d = d, p_re = RePtr, p_im = ImPtr, bfptrre = BfRePtr, bfptrim = BfImPtr;
    	     k ; k--, p_re++, p_im++ )
    	{
    	    TmpRe   = *bfptrre * p_re[d] + *bfptrim * p_im[d];
    	    TmpIm   = *bfptrre * p_im[d] - *bfptrim * p_re[d];
            p_re[d] = *p_re - TmpRe;
            p_im[d] = *p_im - TmpIm;
            *p_re   = *p_re + TmpRe;
            *p_im   = *p_im + TmpIm;

    	    if( !(--modulo_d) ) { p_re += d; p_im += d; bfptrre++; bfptrim++; modulo_d = d; }
      	}
    }

    /* Ergebnisvektor umsortieren in Bit-Reversal Order */
    for (d = 0; d < Len; d++)
    {   if ( (k = SpBitPtr[d]) > d )
      	{    TmpRe = RePtr[d];
             TmpIm = ImPtr[d];
             RePtr[d] = RePtr[k];
             ImPtr[d] = ImPtr[k];
             RePtr[k] = TmpRe;
             ImPtr[k] = TmpIm;
      	}
    }


   /* Durch n div */
    for( d = 0; d < Len; d++ )
    {   RePtr[d] /= Len;
        ImPtr[d] /= Len;
    }
}

// end of fft.cc

// test

#ifdef TEST_LEVEL

int main()
{   unsigned int order;
    printf( "\nFFT order: " ); scanf( "%d", &order );

    CFastFourier FFT(order);
    unsigned int len = (unsigned int)ldexp(1.0, order);
    float *re = new float[len];
    float *im = new float[len];

    printf("\nRe: ");
    for( int i = 0; i < len; i++ ) scanf( "%f", &re[i] );
    printf("\nIm: ");
    for( i = 0; i < len; i++ ) scanf( "%f", &im[i] );

    FFT.transform( re, im );

    printf("\n\nRe: ");
    for( i = 0; i < len; i++ ) printf( "%4.1f", re[i] );
    printf("\nIm: ");
    for( i = 0; i < len; i++ ) printf( "%4.1f", im[i] );

    FFT.invers( re, im );

    printf("\n\nRe: ");
    for( i = 0; i < len; i++ ) printf( "%4.1f", re[i] );
    printf("\nIm: ");
    for( i = 0; i < len; i++ ) printf( "%4.1f", im[i] );

    return(0);
}


#endif  // TEST_LEVELcd source
