/******************************************************************************
 * file:        pref_drs.cc  
 * info:        implementation of base class CSpecConv and CPreFeature
 *              spectral convolution for bark scaled spectrum (TU-Dresden)
 * written by:  M.Beham, 12.01.95 (org. melfilt)
 * last change: 
 * 
 ******************************************************************************/
/******************************************************************************
 *
 * Copyright (C) 1995 M. Beham
 *                                                                   
 * Alle Rechte vorbehalten. Diese Programmlistings und das daraus
 * erhaltene Programm duerfen nur mit schriftlicher Genehmigung	
 * des Autors weitergegeben werden.
 * Insbesondere duerfen auch Auszuege und die daraus erhaltenen
 * Programmteile nur mit schriftlicher Genehmigung des Autors
 * weitergeben, geaendert oder in Programmen anderer Autoren
 * verwendet werden.
 * Diese Copyright-Information darf nicht entfernt oder geaendert
 * werden.
 ******************************************************************************/
/******************************************************************************
 *
 * Copyright (C) C.M.Westendorf (TU-Dresden) und SIEMENS AG  15.06.1994
 *
 * Die Implementierung realisiert ein von der SIEMENS AG entwickeltes Signal-
 * analyseverfahren, da"s an der TU-Dresden neu implementiert wurde.
 *
 * Die Software ist ausschlie"slich f"ur die Verwendung innerhalb des Projektes
 * VERBMOBIL bestimmt. Eine Weitergabe der Programme an Partner au"serhalb des 
 * Projektes VERBMOBIL ist nur mit Zustimmung der SIEMENS AG und des Institutes
 * f"ur Technische Akustik der TU Dresden, sowie des Lehrstuhls MMK der TU
 * Muenchen zul"assig. 
 * 
 * Das Projekt VERBMOBIL wird vom Bundesminister f"ur Forschung und Technologie
 * unter dem F"orderkennzeichen 01IV 102K gef"ordert.
 * 
 ******************************************************************************/

#include <math.h>
#include <string.h>

// #define DEBUG_LEVEL 4
// #define TEST_LEVEL 
// #define LOG_RESULTS

#include "dbg_mess.h"
#include "pref_drs.h"

#define  MAX(a,b) ( (a)>(b)? (a) : (b) )

//--------------------------- public 'CSpecConv' ------------------------------

// constructor: CSpecConv
CSpecConv::CSpecConv( unsigned int powerspeclen, unsigned int convlen )
 :  n_in( powerspeclen ),
    n_out( convlen )
{
    DEBUG_1_MSG("'CSpecConv' constructor called" );
    typedef float *TFloatPtr;

    // alloc memory for all tabs
    mid    = new short[n_out+1];
    width  = new short[n_out];
    norm   = new float[n_out];
    a      = new TFloatPtr[n_out+1];
    IF_FATAL_EXIT( mid==0 || width==0 || norm==0 || a==0, "'CSpecConv' constructor out of memory");

    // calc. convolution coeffs
    mid[0] = 3;    
    for( int k=0; k<n_out; k++)
    {   if( k <= 9 )
        {   mid[k+1] = mid[k] + 3;
            width[k] = 24;
        }
        if( ( k > 9 ) && ( k <= 20) )
        {   mid[k+1] = mid[k] + 4;
            width[k] = 32;
        }
        if( k > 20 )
        {   mid[k+1] = mid[k] + (mid[k] - mid[k-1] + 2);
            width[k] = 8 * ( mid[k+1] - mid[k] );
        }

        a[k] = new float[width[k]+1];
        IF_FATAL_EXIT( a[k] == 0, "'CSpecConv' constructor out of memory");

        a[k][0] = 1.0;
        norm[k] = 0.0;

#ifdef LOG_RESULTS
	    printf( "%2d. mid: %3d; width: %3d; Akj: %f", k, mid[k], width[k], a[k][0] );
#endif

        short nz;
        for ( int j=1; j <= width[k]; j++)
        {   nz      =  width[k]/4;
            a[k][j] =  nz * sin ((double)(j*M_PI)/nz) / (j*M_PI);
            norm[k] += a[k][j];
#ifdef LOG_RESULTS
	    printf( " %f", a[k][j] );
#endif
        }
        norm[k] *= 2.0;
        norm[k] += a[k][0];
#ifdef LOG_RESULTS
        printf( "; norm: %f\n", norm[k] );
#endif
    }


}


// destructor: CSpecConv
CSpecConv::~CSpecConv( )
{
    DEBUG_1_MSG("'CSpecConv' destructor called" );

    for( int k = 0; k < n_out; k++ )
        delete [] a[k];
    delete [] a;
    delete [] norm;
    delete [] width;
    delete [] mid;
}

#define   ABSOL(A)    ( (A) > 0 ?  (A) : -(A) )
#define   ACLIP(A,C,N)  ( (A) >= (C) ? (N)-(A) : (A) )

// convolute one power spectrum
void CSpecConv::convSpec( float *powspec, float *convout ) const
{
    DEBUG_4_MSG("'CSpecConv' convSpec called:" );
 
    register  short i,j,k;
    register  float z;
    register  short n2_1 = 2 * n_in - 1;

#ifdef LOG_RESULTS
    double sum = 0.0;
    printf( "power-spec.:" );
    for( i=0; i<n_in; i++ )
	{ printf( " %f", powspec[i] );
	  sum += powspec[i];
	}
    printf( "; sum: %f\n", sum );
#endif

    for( k=0; k < n_out; k++)
    {   
        i    =  mid[k+1];
        z    =  a[k][0] * powspec[i]; 
        for (j=1; j< width[k]; j++)
            z += a[k][j] * (powspec[ABSOL(i-j)] + powspec[ACLIP(i+j,n_in,n2_1)]);

        convout[k] = z / norm[k];
   }


#ifdef LOG_RESULTS
    printf( "conv.-spec.:" );
    for( i=0; i<n_out; i++ )
	printf( " %f", convout[i] );
    printf( "\n" );
#endif

}

//--------------------------- end of 'CSpecConv' ------------------------------

//--------------------------- public 'CSpDetFeature' --------------------------

// constructor: CSpDetFeature
CSpDetFeature::CSpDetFeature( double dt )
:   b( exp(-dt/SPDETDELTATAU) ),
    a( 1.0 - b )
{
    DEBUG_1_MSG("'CSpDetFeature' constructor called" );
    float *yp = (float*)&YVec;
    for( unsigned int i=0; i<sizeof(SSpDetFeature)/sizeof(float); i++ )
        yp[i] = 0.0;
}

// destructor
CSpDetFeature::~CSpDetFeature( )
{
    DEBUG_1_MSG("'CSpDetFeature' destructor called" );
}

// clculate one feature-vector for speech non-speech detector
float *CSpDetFeature::getFeature( const SPreFeature *pfptr )
{
    DEBUG_4_MSG("'CSpDetFeature' getFeature called" );
    register unsigned int i,j;
    register float *x1fp = (float*) &(YVec.Vec);
    register float *yfp  = (float*) &(YVec.DVec);
    register const float *xfp;

    // reduce pre.feature vector 
    XVec.TotalEnergy = pfptr->TotalEnergy;
    XVec.ZCRate      = pfptr->ZCRate;
    XVec.MidSpec     = pfptr->MidSpec;
    for( xfp=pfptr->Spec, i=0; i<SPDETSPECLEN; i++ )
        for( j=1, XVec.SmSpec[i] = *xfp++; j<SPDETSPECRED; j++ )
            XVec.SmSpec[i] += *xfp++;

    // calc. deltas
    xfp = (float *)&XVec;
    for( i=0; i<sizeof(SSpDetFeatX)/sizeof(float); i++ )
        yfp[i] = b * yfp[i] + a * (xfp[i]-x1fp[i]);
    memcpy( x1fp, xfp, sizeof(SSpDetFeatX) );

    return x1fp;
}


//--------------------------- end of 'CSpDetFeature' --------------------------

//--------------------------- public 'CPreFeature' ----------------------------


// constructor: CPreFeature
CPreFeature::CPreFeature( unsigned int powerspeclen, double frqsampl, double noiseDB, double quantDB )
 :  SpecConv( powerspeclen, PRESPECLEN ),
    SpDetFeature( WINDOW_DELTA / frqsampl ),
    PreSpecLen( PRESPECLEN ),
    PowerSpecLen( powerspeclen ),
    QuantLevel( pow(10.0, quantDB/10.0 ) ),
    NoisePower( new float[powerspeclen] )
{
    DEBUG_1_MSG("'CPreFeature' constructor called" );
    IF_FATAL_EXIT( NoisePower == 0, "'CSpecConv' constructor out of memory");

    double np = (pow( 10.0, noiseDB/10.0 ) * powerspeclen) / 4.0;
    for( register int i=0; i<PowerSpecLen; i++ )
	NoisePower[i] = np;
}

// destructor: CPreFeature
CPreFeature::~CPreFeature( )
{
    DEBUG_1_MSG("'CPreFeature' destructor called" );
    delete NoisePower;
}

static const double Log10Norm = log(10.0) / 10.0;

// convolute one power spectrum to pre. log. spectrum
void CPreFeature::convPowSpec( float *powspecptr, float zcr, SPreFeature *pfptr ) const
{
    register int i;
    double sum = 0.0;
    float  normF;

    DEBUG_4_MSG("'CPreFeature' convSigAnl called:" );

    // subtract Noise and clip to QuantLevel
    for( i=0; i<PowerSpecLen; i++ )
	powspecptr[i] = MAX( powspecptr[i]-NoisePower[i], QuantLevel );
 
    // calc total energy
    for( i=0; i<PowerSpecLen; i++ )
	sum += powspecptr[i];
    pfptr->TotalEnergy = 10.0*log10(sum * (4.0/(PowerSpecLen*PowerSpecLen)));

#ifdef LOG_RESULTS
    printf( "  lin-spec.:" );
    for( i = 0; i < PowerSpecLen; i++ )
       printf( " %f", powspecptr[i] );
    printf( "; total: %f dB\n", pfptr->TotalEnergy );
#endif
    // log. power-spectrum and normalize Power-Spec.
    for( i=0; i<PowerSpecLen; i++ )
	powspecptr[i] = 10.0*log10( powspecptr[i] );

    // conv. power-spectrum, no normalization
    SpecConv.convSpec( powspecptr, pfptr->Spec );
    
    // normalize conv.Spec to total-energy
    sum = 0.0;
    for( i=0; i<PreSpecLen; i++ )
	sum += exp( pfptr->Spec[i] * Log10Norm );
    normF = 10.0*log10( sum / PreSpecLen );
    for( i=0; i<PreSpecLen; i++ )
	pfptr->Spec[i] -= normF;
    pfptr->MidSpec = normF;
    // ZCR
    pfptr->ZCRate = zcr / (2*PowerSpecLen);
}

// calculate additional features  (post proc.)
void CPreFeature::postPreSpec( SPreFeature *pfptr ) const
{
    DEBUG_4_MSG("'CPreFeature' postPreSpec called:" );
    pfptr = 0;
}

// norm factor for signal window
double CPreFeature::getNormFactor() const
{
    return 1.0;
}

// print one pre.feature vector
void  CPreFeature::printPreFeature( SPreFeature *pfptr ) const
{
    printf( "%4.1f %4.2f %4.1f |", pfptr->TotalEnergy, pfptr->ZCRate, pfptr->MidSpec );
    for( int i=0; i<PreSpecLen; i++ )
        printf( " %4.1f", pfptr->Spec[i] );
}

// print one delta feature vector
void  CPreFeature::printDeltaFeature( SDeltaFeature *pfptr ) const
{
    printf( "%4.1f %4.2f %4.1f |", pfptr->TotalEnergy, pfptr->ZCRate, pfptr->MidSpec );
    for( int i=0; i<DELTASPECLEN; i++ )
        printf( " %4.1f", pfptr->Spec[i] );
}

// print one pre.feature vector
void  CPreFeature::printFeature( SFeature *fptr ) const
{
    printPreFeature( &fptr->Pre );
    printf( "\n" );
    printDeltaFeature( &fptr->Delta );
    printf( "\n" );
    printDeltaFeature( &fptr->DDelta );
    printf( "\n" );
}


//--------------------------- end of 'CPreFeature' ----------------------------

//--------------------------- public 'CDeltaFeature' --------------------------

// constructor: CDeltaFeature
CDeltaFeature::CDeltaFeature( unsigned int prespeclen, unsigned int deltaspeclen )
:   PreSpecLen(   prespeclen ),
    DeltaSpecLen( deltaspeclen )
{
    DEBUG_1_MSG("'CDeltaFeature' constructor called" );
    SpecMerge = 2;
    IF_FATAL_EXIT( SpecMerge*DeltaSpecLen != PreSpecLen, "'CDeltaFeature' spec.length != 2 * deltaspec.length");
}

// destructor: CDeltaFeature
CDeltaFeature::~CDeltaFeature( )
{
    DEBUG_1_MSG("'CPreFeature' destructor called" );
}

// calculate delta feature
void CDeltaFeature::calcDelta( SFeatBufEntry *ep ) const
{
    DEBUG_4_MSG("'CDeltaFeature' calcDelta called" );

    // calc deltas
    ep->Vec.Delta.TotalEnergy = ep->Vec.Pre.TotalEnergy - ep->DeltaBack->Vec.Pre.TotalEnergy;
    ep->Vec.Delta.ZCRate      = ep->Vec.Pre.ZCRate      - ep->DeltaBack->Vec.Pre.ZCRate;
    ep->Vec.Delta.MidSpec     = ep->Vec.Pre.MidSpec     - ep->DeltaBack->Vec.Pre.MidSpec;
    // Down sample spec and deltacalc
    for( int j=0, i=0; i<DeltaSpecLen; i++, j+=2 )
         ep->Vec.Delta.Spec[i] = 
             (ep->Vec.Pre.Spec[j]            + ep->Vec.Pre.Spec[j+1] - 
              ep->DeltaBack->Vec.Pre.Spec[j] - ep->DeltaBack->Vec.Pre.Spec[j+1]) / 2.0;
}

// calculate delta delta feature
void CDeltaFeature::calcDDelta( SFeatBufEntry *ep ) const
{
    DEBUG_4_MSG("'CDeltaFeature' calcDDelta called" );

    // calc delta-deltas
    ep->Vec.DDelta.TotalEnergy = ep->DeltaForw->Vec.Delta.TotalEnergy - ep->Vec.Delta.TotalEnergy;
    ep->Vec.DDelta.ZCRate      = ep->DeltaForw->Vec.Delta.ZCRate      - ep->Vec.Delta.ZCRate;
    ep->Vec.DDelta.MidSpec     = ep->DeltaForw->Vec.Delta.MidSpec     - ep->Vec.Delta.MidSpec;
    for( int i=0; i<DeltaSpecLen; i++ )
         ep->Vec.DDelta.Spec[i] = ep->DeltaForw->Vec.Delta.Spec[i] - ep->Vec.Delta.Spec[i];
}

//--------------------------- end of 'CDeltaFeature' --------------------------

//--------------------------- public 'CGainControl ----------------------------

// constructor
CGainControl::CGainControl( FILE *fp, double dt )
:   TauRise( "GainCtrl.TauRise", fp, 1.0 ),
    TauFall( "GainCtrl.TauFall", fp, 1.0 ),
    UpdateUse( "GainCtrl.UpdateUse", ESPEECHFLAG, fp, S ),
    UpdateFlag( "GainCtrl.UpdateMax", fp,  FALSE ),
    UpdateWrite( "GainCtrl.UpdateWrite", fp, FALSE ),
    Max( "GainCtrl.ShortTimeMax", sizeof(SNormFeature)/sizeof(float), fp, 0.0 ),
    bRise( (float)exp(-dt/TauRise.Par) ),
    bFall( (float)exp(-dt/TauFall.Par) ),
    aRise( 1.0 - bRise ),
    aFall( 1.0 - bFall ),
    ParFile( fp ),
    EnergyDelta( "GainCtrl.EnergyDelta", fp, 40.0 )
{
    DEBUG_1_MSG("'CGainControl' constructor called" );
}

// destructor
CGainControl::~CGainControl()
{
    DEBUG_1_MSG("'CGainControl' destructor called" );
    if( UpdateWrite.Par )
        Max.writePar( ParFile );
}

// destructor
int CGainControl::writePar( FILE *fp )
{
    DEBUG_1_MSG("'CGainControl' writePar called" );
    return Max.writePar( fp );
}

// normalize features and update Max
void CGainControl::norm( SPreFeature *prefp, SNormFeature *normfp, ESpeechFlag speechflag )
{
    DEBUG_4_MSG("'CGainControl' norm called" );

    SNormFeature  *actMax = (SNormFeature *)Max.Vec;
    float         *XVec   = (float *)prefp;

    // normalize features 
    normfp->TotalEnergy = (prefp->TotalEnergy - actMax->TotalEnergy) / EnergyDelta.Par + 1.0;
    if( normfp->TotalEnergy < 0.0 ) normfp->TotalEnergy = 0.0;
    normfp->ZCRate = prefp->ZCRate / actMax->ZCRate;
    normfp->MidSpec = prefp->MidSpec / actMax->MidSpec;

    // update max only if speech flag
    if( UpdateFlag.Par && speechflag >= UpdateUse.Par )
    {    for( unsigned int i=0; i < sizeof(SNormFeature)/sizeof(float); i++ )
             if( XVec[i] > Max.Vec[i] )
                 Max.Vec[i] = bRise * Max.Vec[i] + aRise * XVec[i];
             else
                 Max.Vec[i] = bFall * Max.Vec[i] + aFall * XVec[i];
    }
}

//--------------------------- end of 'CGainControl ----------------------------

//--------------------------- test --------------------------------------------

#ifdef TEST_LEVEL

#define LEN 256

int main( int argc, char *argv[] )
{   
    IF_ERROR_RETURN( argc != 1, "usage: test " );  

    const CPreFeature    ap( LEN, 16000.0 );
    float          power[LEN];
    SPreFeature    fv;

    float a = pow(2.0,15.0)*LEN;

    for(int i=0; i<LEN; i++)
        power[i] = (i == LEN/8) ? a*a : 0.0;

    ap.convPowSpec(power, 0.5, &fv);
    ap.postPreSpec( &fv );

    ap.printPreFeature( &fv ); printf("\n");

    return(0);
}




#endif  // TEST_LEVEL
