/******************************************************************************
 * file:        pref_muc.cc  
 * info:        implementation of base class CSpecConv and CPreFeature
 *              spectral convolution for bark scaled spectrum (TU Muenchen)
 * written by:  M.Beham, 12.01.95 (org. vor)
 * last change: 
 * 
 ******************************************************************************/
/******************************************************************************
 *
 * Copyright (C) 1995 M. Beham
 *                                                                   
 * Alle Rechte vorbehalten. Diese Programmlistings und das daraus
 * erhaltene Programm duerfen nur mit schriftlicher Genehmigung	
 * des Autors weitergegeben werden.
 * Insbesondere duerfen auch Auszuege und die daraus erhaltenen
 * Programmteile nur mit schriftlicher Genehmigung des Autors
 * weitergeben, geaendert oder in Programmen anderer Autoren
 * verwendet werden.
 * Diese Copyright-Information darf nicht entfernt oder geaendert
 * werden.
 ******************************************************************************/
/* 
*/

#include <math.h>
#include <string.h>

// #define DEBUG_LEVEL 4
// #define TEST_LEVEL 

#include "dbg_mess.h"
#include "pref_muc.h"


//--------------------------- public 'CSpecConv' ------------------------------

// constructor: CSpecConv
CSpecConv::CSpecConv( unsigned int powerspeclen, unsigned int convlen, double frqsampl,
                      double firstBark, double deltaBark, double minLimit )
:   DeltaFrqSpec( frqsampl / (2.0*powerspeclen) ),
    PowerSpecLen( powerspeclen ),
    ConvLen( convlen ),
    FirstBarkGroup( firstBark ),
    DeltaBarkGroup( deltaBark ),
    MinWeightLimit( minLimit  )
{
    DEBUG_1_MSG("'CSpecConv' constructor called" );

    // generate matrix for convolution (row by row)
    int    coeffanz, coeffall = 0;
    float *tmpcoeff = new float[PowerSpecLen];
    ConvRowTab      = new SConvCoeffRow[ConvLen];
    IF_FATAL_EXIT( tmpcoeff == 0 || ConvRowTab == 0, "'CSpecConv' constructor out of memory");

    double z = FirstBarkGroup;
    for( int i = 0; i < ConvLen; i++, z+=DeltaBarkGroup )
    {   ConvRowTab[i].MidFrqHz = freq(z);
        coeffanz = coeffrow( tmpcoeff, ConvRowTab[i].MidFrqHz, &ConvRowTab[i].FirstFrq, &ConvRowTab[i].LastFrq );
        if( coeffanz == 0 )
        {   WARNING_P("'CSpecConv' constructor: ");
            fprintf( stderr, "row %d (z=%1.1lf Bark) in conv.matrix is empty\n", i, z );
            ConvRowTab[i].CoeffTab = 0;
        } else
        {   ConvRowTab[i].CoeffTab = new float[coeffanz];
            IF_FATAL_EXIT( ConvRowTab[i].CoeffTab == 0, "'CSpecConv' constructor out of memory");
            memcpy( ConvRowTab[i].CoeffTab, tmpcoeff, coeffanz*sizeof(float) );
        }
        coeffall += coeffanz;
    }

    delete [] tmpcoeff;

#if defined(DEBUG_LEVEL) && (DEBUG_LEVEL >= 2 || DEBUG_LEVEL == -2)
fprintf( stderr, ">> 'CSpecConv' spectral conv. matrix (%dx%d) info:\n", PowerSpecLen, ConvLen );
fprintf( stderr, "  zo = %4.2f Bark, delta-z = %4.2f Bark, delta-f (FFT) = %4.2f Hz, density = %3.1f %%\n",
         FirstBarkGroup, DeltaBarkGroup, DeltaFrqSpec, 100.0*((double)coeffall/(PowerSpecLen*ConvLen)) );
#endif
}

// destructor: CSpecConv
CSpecConv::~CSpecConv( )
{
    DEBUG_1_MSG("'CSpecConv' destructor called" );

    for( int i = 0; i < ConvLen; i++ )
        if( ConvRowTab[i].CoeffTab != 0 ) delete [] ConvRowTab[i].CoeffTab;
    delete [] ConvRowTab;
}

// convolute one power spectrum
void CSpecConv::convSpec( float *powspecptr, float *convoutptr ) const
{
    DEBUG_4_MSG("'CSpecConv' convSpec called:" );
 
    register int    i, j;
    register float *coeffptr;

    for( j=0;  j < ConvLen; j++, convoutptr++ )
    {
    	*convoutptr = 0.0;
        for( coeffptr=ConvRowTab[j].CoeffTab, i=ConvRowTab[j].FirstFrq; i <= ConvRowTab[j].LastFrq; i++ )
    	    *convoutptr += powspecptr[i] * *coeffptr++;
    }
}

//--------------------------- private derived class 'CSpecConv' ---------------

// Tonheit z (Bark) als Funktion der Frequenz (Hz) nach Zwicker
double  CSpecConv::bark( double f)
{
    f *= 1E-3;  /* f in kHz umrechnen */
    return( 13.0*atan(0.76*f) + 3.5*atan((f/7.5)*(f/7.5)) );
}

// Erregungspegel als Funktion der Tonheit
double  CSpecConv::erpeg( double z )
{
enum tab_index_2 { Z, Y };
/* 8 Punkte Annaeherung */
#define PUNKTE	8
    static double  tab[PUNKTE][2] = {
    {-3.0,-12.0 },
    {-0.4, 58.0 },
    {-0.2, 60.0 },
    { 0.2, 60.0 },
    { 2.0, 39.0 },
    { 4.0, 20.0 },
    { 6.0,  6.0 },
    { 9.0, -9.0 }  };
    static double  tabmax = 60.0;

    int  il = 0, ir = PUNKTE-1, i;

    /* Intervall-Suche */
    while( ir - il > 1 )
    {
    	i = (ir + il) / 2;
    	if( tab[i][Z] > z )
    	    ir = i;
    	else
    	    il = i;
    } 
    /* Gerade zwischen den beiden Punkten il, ir */
    return(  ( tab[il][Y] * (z - tab[ir][Z]) - tab[ir][Y] * (z - tab[il][Z]) ) /
             ( tab[il][Z] - tab[ir][Z] ) - tabmax  );
}


// Exponentialfunktion zur Basis 10
double  CSpecConv::exp10( double x )
{
    static double ln10 = 2.30258509299404567;  /* = ln 10 */
    return( exp( x * ln10 ) );
}

// Umrechnen der Tonheit z (Bark) in die Frequenz f (Hz)
double  CSpecConv::freq( double z )
{
    double fm, fl = 0.0, fr = 25000.0;  /* alle Frq. in Hz */

    if( bark( fl ) > z || bark( fr ) < z ) return( -1.0 );

    while( fr - fl > 0.1 )
    {
    	fm = ( fr + fl ) / 2.0;
    	if( bark( fm ) > z )
            fr = fm;
    	else
    	    fl = fm;
    }
    return( (fr + fl) / 2.0 );
}


// Schar der Erregungsverteilungen ueber linearer Frequenzachse: Le(f, v)
double  CSpecConv::erreg( double f, double v )
{
    return(  erpeg( bark(f) - bark(v) )  );
}

// Naeherungsweise Parameter-Integral Berechnung ueber Le(f, v):
double  CSpecConv::coeff( double fj, int i )
{
    const int  K = 100;
    register int k;
    double  v  = DeltaFrqSpec * i - 0.5 * DeltaFrqSpec, sum = 0.0;
    double  dv = DeltaFrqSpec / K;

    for( k=0; k < K; k++, v+=dv )
    	sum += exp10( 0.1 * erreg( fj, v ) );

    return( sum / K );
}

// Eine komplette Zeile der Conv Matrix berechnen:
int  CSpecConv::coeffrow( float *coeffptr, double fj, unsigned short *von, unsigned short *bis )
{
    int     i, flag = TRUE;
    double  g;

    for( i=1; i < PowerSpecLen; i++ )
    {
    	g = coeff( fj, i );
    	if( g > MinWeightLimit )
    	{
    	    if( flag )  { flag = FALSE; *von = i; }
    	    *coeffptr++ = g;
    	}
    	else
    	    if( !flag ) break;
    }
    *bis = i - 1;
    
    if( flag ) return 0;
    else       return *bis - *von + 1;

}

//--------------------------- end of 'CSpecConv' ------------------------------

//--------------------------- public 'CPreFeature' ----------------------------

// constructor: CPreFeature
CPreFeature::CPreFeature( unsigned int powerspeclen, double frqsampl )
:   SpecConv( powerspeclen, PRESPECLEN, frqsampl, FIRSTBARK, DELTABARK, MINLIMIT ),
    PreSpecLen( PRESPECLEN ),
    PowerSpecLen( powerspeclen )
{
    DEBUG_1_MSG("'CPreFeature' constructor called" );
}

// destructor: CPreFeature
CPreFeature::~CPreFeature( )
{
    DEBUG_1_MSG("'CPreFeature' destructor called" );
}

// convolute one power spectrum to pre. log. spectrum
void CPreFeature::convPowSpec( float *powspecptr, float zcr, SPreFeature *pfptr ) const
{
    DEBUG_4_MSG("'CPreFeature' convPowSpec called:" );
 
    // conv. power-spectrum
    SpecConv.convSpec( powspecptr, pfptr->Spec );
    // ZCR
    pfptr->ZCRate = zcr;
    // entropy
    pfptr->ModLoudTimesZCRate = 0.0 ; // just dummy here, overwrite it in main-prog
    pfptr->PZMEnergy = 0.0 ;          // just dummy here, overwrite it in main-prog
    // calc spec. loudness of spec.
    spezloudness( pfptr->Spec );
    // calc Sum. and Mod. Loudness
    pfptr->SumLoud = sumloudness( pfptr->Spec );
    pfptr->ModLoud = modloudness( pfptr->Spec );
    // calc. log(spec). for channel comp.
    logspec( pfptr->Spec );
    // calc total energy
    pfptr->TotalEnergy = totalenergy( powspecptr ) - ENERGYLEVEL;
    if( pfptr->TotalEnergy < 0.0 ) pfptr->TotalEnergy = 0.0;
}



// post proc. of prefeature after channel comp.
void CPreFeature::postPreSpec( SPreFeature *pfptr ) const
{
    DEBUG_4_MSG("'CPreFeature' postPreSpec called:" );
 
    // calc. loudness from log. spectrum
    expspec( pfptr->Spec );
    // normalize spectrum to sum=1
    normalizeSpec( pfptr->Spec );
}

// norm factor for signal window
double CPreFeature::getNormFactor() const
{
    return 1.0 / ldexp( PowerSpecLen*2.0, 16-2 );
}

// print one pre.feature vector
void  CPreFeature::printPreFeature( SPreFeature *pfptr ) const
{
    printf( "%4.1f %4.0f %4.2f %4.2f |", pfptr->TotalEnergy, pfptr->ZCRate, pfptr->SumLoud, pfptr->ModLoud );
    for( int i=0; i<PreSpecLen; i++ )
        printf( " %5.3f", pfptr->Spec[i] );
}

// print one delta feature vector
void  CPreFeature::printDeltaFeature( SDeltaFeature *pfptr ) const
{
    printf( "%4.1f %4.0f %4.2f %4.2f |", pfptr->TotalEnergy, pfptr->ZCRate, pfptr->SumLoud, pfptr->ModLoud );
    for( int i=0; i<DELTASPECLEN; i++ )
        printf( " %5.3f", pfptr->Spec[i] );
}

// print one feature vector
void  CPreFeature::printFeature( SFeature *fptr ) const
{
    printPreFeature( &fptr->Pre );
    printf( "\n" );
    printDeltaFeature( &fptr->Delta );
    printf( "\n" );
    printDeltaFeature( &fptr->DDelta );
    printf( "\n" );
}


//--------------------------- private 'CPreFeature' ---------------------------

// total energy from power spectrum
double  CPreFeature::totalenergy( float *powspec ) const
{
    double  sum = 0.0;
    for( register int i = 0; i < PowerSpecLen; i++ )
    	sum += *powspec++;

    return 10.0 * ( (sum < 1e-30)? -30.0 : log10(sum) );
}

// spez loudness of  pre. spec
void CPreFeature::spezloudness( float *spec ) const
{
    register int i;

    for( i=0; i<PreSpecLen; i++ )
	 spec[i] = (float) pow( (double)spec[i], 0.25 );
}

// log of pre. spec for channel comp.
void CPreFeature::logspec( float *spec ) const
{
    register int i;
    for( i=0; i<PreSpecLen; i++ )
    	spec[i] =  spec[i] < 1E-30 ? log(1E-30) : log( (double)spec[i] );
}

// exp of pre. spec for channel comp.
void CPreFeature::expspec( float *spec ) const
{
    register int i;
    for( i=0; i<PreSpecLen; i++ )
    	spec[i] = (float)exp( (double)spec[i] );
}


// sum of loudness
double CPreFeature::sumloudness( float *spec ) const
{
    register int i;
    double   SumLoud = 0.0;
    for( i=0; i<PreSpecLen; i++ )
        SumLoud += spec[i];

    return SumLoud;
}
// mod loudness
double CPreFeature::modloudness( float *spec )  const
{
    register int i;
    double ModLoud = 0.0;
    for( i=MOD_POS_VON; i<=MOD_POS_BIS && i<PreSpecLen; i++ )
         ModLoud += spec[i];
    for( i=MOD_NEG_VON; i<=MOD_NEG_BIS && i<PreSpecLen; i++ )
         ModLoud -= spec[i];
    return( ModLoud > 0.0 ? ModLoud : 0.0 ) ;
}

// normalize spectrum to sum=1
void CPreFeature::normalizeSpec( float *spec ) const
{
    register int i;
    float        sum = 0.0;
    for( i=0; i<PreSpecLen; i++ )
        sum += spec[i];
    for( i=0; i<PreSpecLen; i++ )
        spec[i] /= sum;
}



// negative entropy from power spectrum
double  CPreFeature::entropy( float *powspec ) const
{
  register int i ;
  float  sum = 0.0;
  float *pdf = new float[PowerSpecLen] ;
  float *weight = new float[PowerSpecLen] ;
  for( i=0 ; i < PowerSpecLen; i++ ) {
    if( i<4 )
      weight[i] = 0.0 ;
    else if( i<10 )
      weight[i] = 0.02 ; // (1/50) 
    else if( i<15 )
      weight[i] = 0.01 ;
    else if( i<20 )
      weight[i] = 0.005 ;
    else if( i<50 )
      weight[i] = 0.001 ;
    else if( i<100 )
      weight[i] = 0.0001 ; // (1/1000)
    else
      weight[i] = 0.0 ;
  } ;


  // first calc sum for converting spec(i) intp pdf(i)
  for( i=0 ; i < PowerSpecLen; i++ )
    sum += *powspec++;

  for( i=0 ; i<PowerSpecLen; i++ )
    pdf[i] = (powspec[i]/sum) ;

  // calc entropy
  sum = 0.0 ;
  for( i=0 ; i<PowerSpecLen; i++ )
    sum -= ( (pdf[i]>1.0e-2 && pdf[i]<0.9) ? (weight[i] * pdf[i] * log(pdf[i])) : 0.0 ) ;

  delete[] pdf ;
  delete[] weight ;

  return sum;
}



//--------------------------- end of 'CPreFeature' ----------------------------

//--------------------------- public 'CDeltaFeature' --------------------------

// constructor: CDeltaFeature
CDeltaFeature::CDeltaFeature( unsigned int prespeclen, unsigned int deltaspeclen )
:   PreSpecLen(   prespeclen ),
    DeltaSpecLen( deltaspeclen )
{
    DEBUG_1_MSG("'CDeltaFeature' constructor called" );
    IF_FATAL_EXIT( DeltaSpecLen != PreSpecLen, "'CDeltaFeature' spec.length != deltaspec.length");
}

// destructor: CDeltaFeature
CDeltaFeature::~CDeltaFeature( )
{
    DEBUG_1_MSG("'CPreFeature' destructor called" );
}

// calculate delta feature
void CDeltaFeature::calcDelta( SFeatBufEntry *ep ) const
{
    DEBUG_4_MSG("'CDeltaFeature' calcDelta called" );
    register int k, i;
    register SFeatBufEntry *np;
    register float fac;

    // calc deltas
    ep->Vec.Delta.TotalEnergy = 0.0;
    ep->Vec.Delta.ZCRate      = 0.0;
    ep->Vec.Delta.SumLoud     = 0.0;
    ep->Vec.Delta.ModLoud     = 0.0;
    for( i=0; i<DeltaSpecLen; i++ )
        ep->Vec.Delta.Spec[i] = 0.0;
    ep->Vec.DNorm.TotalEnergy = 0.0;
    ep->Vec.DNorm.ZCRate      = 0.0;
    ep->Vec.DNorm.SumLoud     = 0.0;
    ep->Vec.DNorm.ModLoud     = 0.0;

    for( np=ep->DeltaBack, k=-DELTA_BACK; k<=DELTA_FORW; k++, np=np->Next ) 
    {   if( k == 0 ) continue;
        fac = (float)k / (float)DELTA_FORW;
        ep->Vec.Delta.TotalEnergy += fac * np->Vec.Pre.TotalEnergy;
        ep->Vec.Delta.ZCRate      += fac * np->Vec.Pre.ZCRate;
        ep->Vec.Delta.SumLoud     += fac * np->Vec.Pre.SumLoud;
        ep->Vec.Delta.ModLoud     += fac * np->Vec.Pre.ModLoud;
        // calc delta spec
        for( i=0; i<DeltaSpecLen; i++ )
             ep->Vec.Delta.Spec[i] += fac * np->Vec.Pre.Spec[i];
        // calc delta gain-ctrl. features
        ep->Vec.DNorm.TotalEnergy += fac * np->Vec.Norm.TotalEnergy;
        ep->Vec.DNorm.ZCRate      += fac * np->Vec.Norm.ZCRate;
        ep->Vec.DNorm.SumLoud     += fac * np->Vec.Norm.SumLoud;
        ep->Vec.DNorm.ModLoud     += fac * np->Vec.Norm.ModLoud;
    }

}

// calculate delta delta feature
void CDeltaFeature::calcDDelta( SFeatBufEntry *ep ) const
{
    DEBUG_4_MSG("'CDeltaFeature' calcDDelta called" );
    register int k, i;
    register SFeatBufEntry *np;
    register float fac;

    // calc deltas
    ep->Vec.DDelta.TotalEnergy = 0.0;
    ep->Vec.DDelta.ZCRate      = 0.0;
    ep->Vec.DDelta.SumLoud     = 0.0;
    ep->Vec.DDelta.ModLoud     = 0.0;
    for( i=0; i<DeltaSpecLen; i++ )
        ep->Vec.DDelta.Spec[i] = 0.0;
    ep->Vec.DDNorm.TotalEnergy = 0.0;
    ep->Vec.DDNorm.ZCRate      = 0.0;
    ep->Vec.DDNorm.SumLoud     = 0.0;
    ep->Vec.DDNorm.ModLoud     = 0.0;

    for( np=ep->DeltaBack, k=-DELTA_BACK; k<=DELTA_FORW; k++, np=np->Next )
    {   if( k == 0 ) continue;
        fac = (float)k / (float)DELTA_FORW;
        ep->Vec.DDelta.TotalEnergy += fac * np->Vec.Delta.TotalEnergy;
        ep->Vec.DDelta.ZCRate      += fac * np->Vec.Delta.ZCRate;
        ep->Vec.DDelta.SumLoud     += fac * np->Vec.Delta.SumLoud;
        ep->Vec.DDelta.ModLoud     += fac * np->Vec.Delta.ModLoud;
        // calc delta spec
        for( i=0; i<DeltaSpecLen; i++ )
             ep->Vec.DDelta.Spec[i] += fac * np->Vec.Delta.Spec[i];
        // calc delta of gain-ctrl. features
        ep->Vec.DDNorm.TotalEnergy += fac * np->Vec.DNorm.TotalEnergy;
        ep->Vec.DDNorm.ZCRate      += fac * np->Vec.DNorm.ZCRate;
        ep->Vec.DDNorm.SumLoud     += fac * np->Vec.DNorm.SumLoud;
        ep->Vec.DDNorm.ModLoud     += fac * np->Vec.DNorm.ModLoud;

    }
}

//--------------------------- end of 'CDeltaFeature' --------------------------

//--------------------------- public 'CGainControl ----------------------------

// constructor
CGainControl::CGainControl( FILE *fp, double dt )
:   TauRise( "GainCtrl.TauRise", fp, 1.0 ),
    TauFall( "GainCtrl.TauFall", fp, 1.0 ),
    UpdateUse( "GainCtrl.UpdateUse", ESPEECHFLAG, fp, S ),
    UpdateFlag( "GainCtrl.UpdateMax", fp,  FALSE ),
    UpdateWrite( "GainCtrl.UpdateWrite", fp, FALSE ),
    Max( "GainCtrl.ShortTimeMax", sizeof(SNormFeature)/sizeof(float), fp, 0.0 ),
    bRise( (float)exp(-dt/TauRise.Par) ),
    bFall( (float)exp(-dt/TauFall.Par) ),
    aRise( (float)(1.0 - (double)bRise) ),
    aFall( (float)(1.0 - (double)bFall) ),
    ParFile( fp ),
    EnergyDelta( "GainCtrl.EnergyDelta", fp, 40.0 )
{
    DEBUG_1_MSG("'CGainControl' constructor called" );
}

// destructor
CGainControl::~CGainControl()
{
    DEBUG_1_MSG("'CGainControl' destructor called" );
    if( UpdateWrite.Par )
        Max.writePar( ParFile );
}

// save parameter
int CGainControl::writePar( FILE *fp )
{
    DEBUG_1_MSG("'CGainControl' writePar called" );
    return Max.writePar( fp );
}

// append updatable parameters
int CGainControl::appendUpdPar( FILE *fp )
{
    DEBUG_1_MSG("'CGainControl' writePar called" );
    return Max.appendPar( fp );
}

// append fix parameters
int CGainControl::appendFixPar( FILE *fp )
{
    int ret = 0 ;
    DEBUG_1_MSG("'CGainControl' writePar called" );
    ret |= UpdateFlag.appendPar( fp );
    ret |= UpdateWrite.appendPar( fp );
    ret |= UpdateUse.appendPar( fp );
    ret |= EnergyDelta.appendPar( fp );
    ret |= TauRise.appendPar( fp );
    ret |= TauFall.appendPar( fp );

    return( ret == 0 ? ret : -1 ) ;
}

// normalize features and update Max
void CGainControl::norm( SPreFeature *prefp, SNormFeature *normfp, ESpeechFlag speechflag )
{
    DEBUG_4_MSG("'CGainControl' norm called" );

    SNormFeature  *actMax = (SNormFeature *)Max.Vec;
    float         *XVec   = (float *)prefp;

    // normalize features 
    normfp->TotalEnergy = (prefp->TotalEnergy - actMax->TotalEnergy) / EnergyDelta.Par + 1.0;
    if( normfp->TotalEnergy < 0.0 ) normfp->TotalEnergy = 0.0;
    normfp->ZCRate = prefp->ZCRate / actMax->ZCRate;
    normfp->SumLoud = prefp->SumLoud / actMax->SumLoud;
    normfp->ModLoud = prefp->ModLoud / actMax->ModLoud;

    // update max only if speech flag
    if( UpdateFlag.Par && speechflag >= UpdateUse.Par )
    {    for( unsigned int i=0; i < sizeof(SNormFeature)/sizeof(float); i++ )
             if( XVec[i] > Max.Vec[i] )
                 Max.Vec[i] = bRise * Max.Vec[i] + aRise * XVec[i];
             else
                 Max.Vec[i] = bFall * Max.Vec[i] + aFall * XVec[i];
    }
}

//--------------------------- end of 'CGainControl ----------------------------


//--------------------------- test --------------------------------------------

#ifdef TEST_LEVEL

#define LEN 128

int main( int argc, char *argv[] )
{   
    IF_ERROR_RETURN( argc != 1, "usage: test " );  

    const CPreFeature    ap( LEN, 16000.0 );
    float          pow[LEN];
    SPreFeature    fv;

    for(int i=0; i<LEN; i++)
        pow[i] = (i == LEN/8) ? 1.0 : 0.0;

    ap.convPowSpec(pow, 0.5, &fv);
    ap.postPreSpec( &fv );

    for( i=0; i<ap.getSpecLen(); i++ )
        printf( " %2d. %5.3e\n", i, fv.Spec[i] );

    printf( " energy = %5.3f, zcr = %5.3f, sumLoud = %5.3f, modLoud = %5.3f\n", fv.TotalEnergy, fv.ZCRate, fv.SumLoud, fv.ModLoud );

    return(0);
}




#endif  // TEST_LEVEL
