/******************************************************************************
 * file:        siginput.cc  
 * info:        implementation of base class CSigInput and derived classes
 *              device driver for sample input
 * written by:  M.Beham, 12.01.95 (org. readsig.c onl.c)
 * last change: 
 * 
 ******************************************************************************/
/******************************************************************************
 *
 * Copyright (C) 1995 M. Beham
 *                                                                   
 * Alle Rechte vorbehalten. Diese Programmlistings und das daraus
 * erhaltene Programm duerfen nur mit schriftlicher Genehmigung	
 * des Autors weitergegeben werden.
 * Insbesondere duerfen auch Auszuege und die daraus erhaltenen
 * Programmteile nur mit schriftlicher Genehmigung des Autors
 * weitergeben, geaendert oder in Programmen anderer Autoren
 * verwendet werden.
 * Diese Copyright-Information darf nicht entfernt oder geaendert
 * werden.
 * 
 ******************************************************************************/
/* 
contens:
    class / dev.driver:   device-str:  comment:
    CSigInput              -           base class
    CSigInputFileILS      FileILS      read ILS-File with PHONDAT Header    
    CSigInputSunADC       SunADC       start extra ADC-process and read pipe
    CSigInputAudioFile    AudioFile    SUN audio-file with header
*/


#include <unistd.h>
#include <fcntl.h> 
#include <stdlib.h>
#include <signal.h>
//#include <sys/audioio.h>

// #define DEBUG_LEVEL (-3)
// #define TEST_LEVEL 3
// #define PRG_SWAPILS
// #define PRG_ILS_SUN

// R.Faltlhauser: Modification for LINUX
#ifdef LINUX
#undef RealTimeADC
#include "audioio.h"                // local copy of SUN-original for LINUX (missing under LINUX)
#else
#include <sys/audioio.h>
#endif

//#define RealTimeADC 

#include "dbg_mess.h"
#include "siginput.h"

// const. for FileILS
#define  BL_PER_REQ_FILEILS  2
#define  BL_PER_REQ_FILENIST  2

// const. for SunADC
#define  PROG_SUNADC         "SunADC"
#define  BL_PER_REQ_SUNADC   2
#define  TIMEOUT_SUNADC      10
#define  TRYREAD_SUNADC      100

// const. for AudioFile
#define  BL_PER_REQ_AUDIOFILE      2
#define  CHANNELS_AUDIOFILE        1
#define  ENCODING_AUDIOFILE        AUDIO_ENCODING_LINEAR


//--------------------------- public base class 'CSigInput' -----------------

// constructor base class: CSigInput
CSigInput::CSigInput( unsigned int blocksize, int blocksperrequest )
:   BlockSize( blocksize ),
    BlocksPerRequest( blocksperrequest ),
    FileDes(-1),
    FrqSample(0),
    SampleAnz(0)
{
    DEBUG_1_MSG("'CSigInput' base constructor called" );
}

// destructor base class: CSigInput
CSigInput::~CSigInput( )
{
    DEBUG_1_MSG("'CSigInput' base destructor called" );
}


//--------------------------- end of base class 'CSigInput' -------------------

//--------------------------- public derived class 'CSigInputFileILS' ---------

// constructor: CSigInputFileILS
CSigInputFileILS::CSigInputFileILS( unsigned int blocksize )
:   CSigInput(blocksize, BL_PER_REQ_FILEILS)
{
    DEBUG_1_MSG("'CSigInputFileILS' derived constructor called" );
}

// constructor: CSigInputFileILS
CSigInputFileILS::~CSigInputFileILS( )
{
    DEBUG_1_MSG("'CSigInputFileILS' derived destructor called" );
    
    if( FileDes >= 0 ) close( FileDes );
}

// open device ILS-File and read header
int CSigInputFileILS::startRead( const char *filename, int NrSkip )
{
    if( filename == 0 )
        filename = "";
    DEBUG_2_MSG_S("'CSigInputFileILS' startRead called:", filename );

    // close open stream
    if( FileDes >= 0 ) 
    {   close( FileDes );
        FileDes = -1;
        FrqSample = 0;
    }


    // open file (or stdin)
    if( *filename != '\0' )
    {   for( int i=0; i<3 && (FileDes = open( filename, O_RDONLY )) < 0; i++ ) ;
        IF_ERROR_RETURN_PERR_S( FileDes < 0, "'CSigInputFileILS' startRead: can't open file\n", filename );
    } else
        FileDes = 0;

    // header read and test
    if( read( FileDes, &Header, sizeof(Header) ) != sizeof(Header) )
    {   ERROR_MSG_S( "'CSigInputFileILS' startRead can't read header from:", filename );
        close( FileDes );
        FileDes = -1;
        return _ERR_;
    }
    // PHONDAT normal
    if( Header.flag_type == 0xFFFF8300 || Header.flag_type == 0x00008300 )
        ByteSwap = FALSE;
    else if( Header.flag_type == 0x0083FFFF || Header.flag_type == 0x00830000 )
        ByteSwap = TRUE;
    else
    {   ERROR_MSG_S( "'CSigInputFileILS' startRead no ILS-header in:", filename );
        close( FileDes );
        FileDes = -1;
        return _ERR_;
    }
 
    // header byte swap
    if( ByteSwap )
        header_swap();
#ifdef INFO
    fprintf( stderr, "INFO: Anzahl Header: %d, Headersize: %d, Gesamt: %d\n", Header.anz_header, sizeof(Header),   Header.anz_header * sizeof(Header) ) ;
#endif
    // PHONDAT header version = 2
    if( Header.version == 2 )
    {   if( lseek( FileDes, (Header.anz_header-1)*sizeof(Header)+NrSkip, SEEK_CUR ) != Header.anz_header*sizeof(Header)+NrSkip )
        {   ERROR_MSG_S( "'CSigInputFileILS' startRead can't seek ILS-header in:", filename );
            close( FileDes );
            FileDes = -1;
            return _ERR_;
         }
    }
  
    FrqSample = Header.isf;
    SampleAnz = 0;

#if defined(DEBUG_LEVEL) && ( DEBUG_LEVEL>=3 || DEBUG_LEVEL==-3 )
 fprintf( stderr, ">>> ILS-file: %s, HeaderVersion=%d, ByteSwap=%d, FrqSamp=%d, AdcBits=%d, Date=%2d.%2d.%d", filename, Header.version, ByteSwap, FrqSample, Header.adc_bits, (int)Header.date.day, (int)Header.date.month, (int)Header.date.year );
 if( Header.version == 2 )
    fprintf( stderr, ", HeaderAnz=%d\n", Header.anz_header );
 else
    fprintf( stderr, "\n" );
#endif // DEBUG_LEVEL 3
    

    return _OK_;
}

// stop device, close ILS File
int CSigInputFileILS::stopRead( EStopMode mode )
{
    DEBUG_2_MSG("'CSigInputFileILS' stopRead called" );

    // close ILS file (stop and clear)
    if( mode == Stop && FileDes < 0)
    {   WARNING_MSG("'CSigInputFileILS' stopRead: device already stoped" );
        return _ERR_;
    }
    else if( FileDes >= 0 ) 
    {  close(FileDes);
       FileDes = -1;
       FrqSample = 0;
    }

    return _OK_;
}

// read device ILS-file
int CSigInputFileILS::sigReadBlock( short *buf )
{
    DEBUG_4_MSG("'CSigInputFileILS' sigReadBlock called" );

    // read block
    long rcount = read( FileDes, buf, BlocksPerRequest*BlockSize*sizeof(short) );
    long samples = rcount/sizeof(short);
    SampleAnz += samples;

    // byte swap
    if( ByteSwap )
        for( register long i = 0; i < samples; i++, buf++ )
            short_swap( buf );

    if( rcount !=  BlocksPerRequest*BlockSize*sizeof(short) )
        return _ERR_;
    else
        return _OK_;
}

// write block
int CSigInputFileILS::writeBlock( int fd, short *buf )
{   
    // write block
    long wcount = write( fd, buf, BlocksPerRequest*BlockSize*sizeof(short) );
    if( wcount != BlocksPerRequest*BlockSize*sizeof(short) )
    {   ERROR_MSG( "'CSigInputFileILS' writeBlock: write error" );
        return _ERR_;
    }

    return _OK_;
}

// write ils header
int CSigInputFileILS::writeHeader( char *filename )
{   
    // open file
    int fd = open( filename, O_WRONLY | O_CREAT );
    if( fd < 0 )
    {   ERROR_MSG_S("'CSigInputFileILS' writeILSHeader: can't open file:", filename );
        return fd;
    }

    // header write
    int tmp = Header.version;
    Header.version = 0;
    if( write( fd, &Header, sizeof(Header) ) != sizeof(Header) )
    {   ERROR_MSG_S( "'CSigInputFileILS' write ILSHeader: can't write header to:", filename );
        close( fd );
        return -1;
    }
    Header.version = tmp;

    return fd;
}

//--------------------------- private derived class 'CSigInputFileILS' --------

// byte swap routines
void CSigInputFileILS::header_swap()
{
    long_swap(  &Header.nspbk );
    long_swap(  &Header.anz_header );
    long_swap(  &Header.isf );
    long_swap(  &Header.flag_type );
    long_swap(  &Header.flag_init );
    short_swap( &Header.date.year );
    short_swap( &Header.adc_bits );
    short_swap( &Header.words );
    short_swap( &Header.wdh );
    short_swap( &Header.abs_ampl );
}

void CSigInputFileILS::long_swap( long *lp )
{
    register char *byte = (char*)lp;
    register char tmp = byte[3];
    byte[3] = byte[0];
    byte[0] = tmp;
    tmp     = byte[2];
    byte[2] = byte[1];
    byte[1] = tmp;
}

void CSigInputFileILS::short_swap( short *sp )
{
    register char tmp = ((char*)sp)[0];
    ((char*)sp)[0] = ((char*)sp)[1];
    ((char*)sp)[1] = tmp;
}

//--------------------------- end of derived class 'CSigInputFileILS' ---------



//--------------------------- public derived class 'CSigInputFileNIST' ---------

// constructor: CSigInputFileNIST
CSigInputFileNIST::CSigInputFileNIST( unsigned int blocksize )
:   CSigInput(blocksize, BL_PER_REQ_FILENIST)
{
    DEBUG_1_MSG("'CSigInputFileNIST' derived constructor called" );
}

// constructor: CSigInputFileNIST
CSigInputFileNIST::~CSigInputFileNIST( )
{
    DEBUG_1_MSG("'CSigInputFileNIST' derived destructor called" );
    
    if( FileDes >= 0 ) close( FileDes );
}

// open device NIST-File and read header
int CSigInputFileNIST::startRead( const char *filename, int NrSkip )
{
    int ret ;
    FILE *fp ;

    if( filename == 0 )
        filename = "";
    DEBUG_2_MSG_S("'CSigInputFileNIST' startRead called:", filename );

    // close open stream
    if( FileDes >= 0 ) 
    {   close( FileDes );
        FileDes = -1;
        FrqSample = 0;
    }


    // open file (or stdin)
    //if( *filename != '\0' )
    //{   for( int i=0; i<3 && (FileDes = open( filename, O_RDONLY )) < 0; i++ ) ;
    //    IF_ERROR_RETURN_PERR_S( FileDes < 0, "'CSigInputFileNIST' startRead: can't open file", filename );
    //} else
    //    FileDes = 0;

    if( (fp = fopen( filename, "r" )) == 0 ) {
      fprintf( stderr, "'CSigInputFileNIST' startRead: can't open file %s\n", filename );
      return(-1) ;
    } ;

    // read header
    ret = readHeader( fp ) ;

    fclose( fp ) ;


    // open file (or stdin)
    if( *filename != '\0' )
    {   for( int i=0; i<3 && (FileDes = open( filename, O_RDONLY )) < 0; i++ ) ;
        IF_ERROR_RETURN_PERR_S( FileDes < 0, "'CSigInputFileNIST' startRead: can't open file", filename );
    } else
        FileDes = 0;

    // NIST normal
    ByteSwap = FALSE;

    // header byte swap
    if( ByteSwap )
        header_swap();
#ifdef INFO
    fprintf( stderr, "INFO: Anzahl Header: %d, Headersize: %d, Gesamt: %d\n", Header.anz_header, sizeof(Header),   Header.anz_header * sizeof(Header) ) ;
#endif
    
    // jump to first data entry:
    if( lseek( FileDes, Header.size+NrSkip, SEEK_SET ) != Header.size+NrSkip ) {
      fprintf( stderr, "' CSigInputFileNIST::startRead': could not jump to first data-entry\n" ) ;
      close( FileDes ) ;
      FileDes = -1 ;
      return( _ERR_ ) ;
    } ;
 
    FrqSample = Header.samplerate;
    SampleAnz = 0;

#if defined(DEBUG_LEVEL) && ( DEBUG_LEVEL>=3 || DEBUG_LEVEL==-3 )
 fprintf( stderr, ">>> NIST-file: %s, HeaderVersion=%d, ByteSwap=%d, FrqSamp=%d, AdcBits=%d, Date=%2d.%2d.%d", filename, Header.version, ByteSwap, FrqSample, Header.adc_bits, (int)Header.date.day, (int)Header.date.month, (int)Header.date.year );
 if( Header.version == 2 )
    fprintf( stderr, ", HeaderAnz=%d\n", Header.anz_header );
 else
    fprintf( stderr, "\n" );
#endif // DEBUG_LEVEL 3
    

    return _OK_;
}

// stop device, close NIST File
int CSigInputFileNIST::stopRead( EStopMode mode )
{
    DEBUG_2_MSG("'CSigInputFileNIST' stopRead called" );

    // close NIST file (stop and clear)
    if( mode == Stop && FileDes < 0)
    {   WARNING_MSG("'CSigInputFileNIST' stopRead: device already stoped" );
        return _ERR_;
    }
    else if( FileDes >= 0 ) 
    {  close(FileDes);
       FileDes = -1;
       FrqSample = 0;
    }

    return _OK_;
}

// read device NIST-file
int CSigInputFileNIST::sigReadBlock( short *buf )
{
    DEBUG_4_MSG("'CSigInputFileNIST' sigReadBlock called" );

    // read block
    long rcount = read( FileDes, buf, BlocksPerRequest*BlockSize*sizeof(short) );
    long samples = rcount/sizeof(short);
    SampleAnz += samples;

    // byte swap
    if( ByteSwap )
        for( register long i = 0; i < samples; i++, buf++ )
            short_swap( buf );

    if( rcount !=  BlocksPerRequest*BlockSize*sizeof(short) )
        return _ERR_;
    else
        return _OK_;
}

// write block
int CSigInputFileNIST::writeBlock( int fd, short *buf )
{   
    // write block
    long wcount = write( fd, buf, BlocksPerRequest*BlockSize*sizeof(short) );
    if( wcount != BlocksPerRequest*BlockSize*sizeof(short) )
    {   ERROR_MSG( "'CSigInputFileNIST' writeBlock: write error" );
        return _ERR_;
    }

    return _OK_;
}

// write ils header
int CSigInputFileNIST::writeHeader( char *filename )
{   
    // open file
    int fd = open( filename, O_WRONLY | O_CREAT );
    if( fd < 0 )
    {   ERROR_MSG_S("'CSigInputFileNIST' writeNISTHeader: can't open file:", filename );
        return fd;
    }

    // header write
    //int tmp = Header.version;
    //Header.version = 0;
    //if( write( fd, &Header, sizeof(Header) ) != sizeof(Header) )
    //{   ERROR_MSG_S( "'CSigInputFileNIST' write NISTHeader: can't write header to:", filename );
    //    close( fd );
    //    return -1;
    //}
    //Header.version = tmp;

    return fd;
}

//--------------------------- private derived class 'CSigInputFileNIST' --------

// byte swap routines
void CSigInputFileNIST::header_swap()
{
  //long_swap(  &Header.nspbk );
  //long_swap(  &Header.anz_header );
  //long_swap(  &Header.isf );
  //long_swap(  &Header.flag_type );
  //long_swap(  &Header.flag_init );
  //short_swap( &Header.date.year );
  //short_swap( &Header.adc_bits );
  //short_swap( &Header.words );
  //short_swap( &Header.wdh );
  //short_swap( &Header.abs_ampl );
}

void CSigInputFileNIST::long_swap( long *lp )
{
  //register char *byte = (char*)lp;
  //register char tmp = byte[3];
  //byte[3] = byte[0];
  //byte[0] = tmp;
  //tmp     = byte[2];
  //byte[2] = byte[1];
  //byte[1] = tmp;
}

void CSigInputFileNIST::short_swap( short *sp )
{
  //register char tmp = ((char*)sp)[0];
  //((char*)sp)[0] = ((char*)sp)[1];
  //((char*)sp)[1] = tmp;
}

// dummy version of reading NIST-header (just for reading PCM linear)
int CSigInputFileNIST::readHeader( FILE *fp )
{
  char line[MaxLineLength] ;
  int dummy ;

  // check first line for keyword 'NIST'
  if( fgets( line, MaxLineLength, fp ) == 0 ) {
    fprintf( stderr, "'CSigInputFileNIST::readHeader: can't read first line of file\n" ) ;
    return(-1) ;
  } ;
  if( strstr( line, "NIST" ) == 0 ) {
    fprintf( stderr, "'CSigInputFileNIST::readHeader: can'tfind 'NIST' keyword in first line\n" ) ;
    return(-1) ;
  } ;
  
  // read next line (length of header in bytes!)
  fgets( line, MaxLineLength, fp ) ;
  Header.size = strtol( line, NULL, 10 ) ;

  // read ASCII-lines until 'end_head' found
  while( fgets( line, MaxLineLength, fp ) != 0 && strcmp( line, "end_head" ) != 0 ) {
    
    // check for signal-wordsy

    // channel_count
    if( strstr( line, "channel_count" ) != 0 ) {
      strtok( line, " " ) ;
      strtok( NULL, " " ) ;
      Header.channel = strtol( strtok( NULL, "\n " ), NULL, 10 ) ;
      continue ;
    } ;
  
    // sample_coding
    if( strstr( line, "sample_coding" ) != 0 ) {
      strtok( line, " " ) ;
      strtok( NULL, " " ) ; 
      strcpy( Header.samplecod, strtok( NULL, "\n " )) ; 
      continue ;
    } ;  

    //sample_count
    if( strstr( line, "sample_count" ) != 0 ) {
      strtok( line, " " ) ;
      strtok( NULL, " " ) ;
      Header.samples = strtol( strtok( NULL, "\n " ), NULL, 10 ) ;
      continue ;
    } ;
     
    //sample_rate
    if( strstr( line, "sample_rate" ) != 0 ) {
      strtok( line, " " ) ;
      strtok( NULL, " " ) ;
      Header.samplerate = strtol( strtok( NULL, "\n " ), NULL, 10 ) ;
      continue ;
    } ;
    
    //sample_n_bytes
    if( strstr( line, "sample_n_bytes" ) != 0 ) {
      strtok( line, " " ) ;
      strtok( NULL, " " ) ;
      Header.sample_n_bytes = strtol( strtok( NULL, "\n " ), NULL, 10 ) ;
      continue ;
    } ;

    //sample_sig_bits
    if( strstr( line, "sample_sig_bits" ) != 0 ) {
      strtok( line, " " ) ;
      strtok( NULL, " " ) ;
      Header.sample_sig_bits = strtol( strtok( NULL, "\n " ), NULL, 10 ) ;
      continue ;
    } ;
   
    // sample_byte_format
    if( strstr( line, "sample_byte_format" ) != 0 ) {
      strtok( line, " " ) ;
      strtok( NULL, " " ) ; 
      strcpy( Header.byteformat, strtok( NULL, "\n " )) ; 
      continue ;
    } ;  

  } ;

  return(1) ;
} ;


//--------------------------- end of derived class 'CSigInputFileNIST' ---------


#ifdef RealTimeADC
//--------------------------- public derived class 'CSigInputSunADC' ---------

// constructor: CSigInputSunADC
CSigInputSunADC::CSigInputSunADC( unsigned int blocksize, unsigned long frqsample )
:   CSigInput(blocksize, BL_PER_REQ_SUNADC)
{
    DEBUG_1_MSG("'CSigInputSunADC' derived constructor called" );

    // set sampling frq
    FrqSample = frqsample;

    // open pipe for data transfer to SunADC
    int  pipefd[2];
    IF_FATAL_EXIT( pipe( pipefd ) < 0, "'CSigInputSunADC' can't open pipe to SunADC");
    int  rdpipefd = pipefd[0];
    int  wrpipefd = pipefd[1];

    // reset ADC ready flag
    Ready.setADC(FALSE);

    // start ADC process here with fork (process dublication!)
    ChildPid = fork();
    if( ChildPid == 0 )
    {
        // here: child process execution (ADC)

        // close read-end of pipe
        close( rdpipefd );
        // close stdout-fd and use wr-end of pipe as stdout
        close( 1 );
        dup( wrpipefd );

        // construct comandline for adc-programm: 
        // SunADC  <BlockSize> <FrqSample>
        char argc[3][16];
        sprintf( argc[0], PROG_SUNADC );
        sprintf( argc[1], "%d", BlockSize*BlocksPerRequest );
        sprintf( argc[2], "%d", FrqSample );
        // start adc-prg; overload child proc. with execlp( <path>/prg, prg-name, args, ..., NULL)
        execlp( argc[0], argc[0], argc[1], argc[2], NULL );

        // the next lines can't be reached if exec is ok.  
        IF_FATAL_EXIT_S( TRUE, "'CSigInputSunADC' (child) can't execute prg.: ", PROG_SUNADC );
    } else
    {
        // here: parent process execution (CSigInput)

        // timeout for adc-ready 
        for( unsigned int i=0; i<=TIMEOUT_SUNADC; i++ ) 
        {   if( Ready.testADC() ) break; 
            sleep( 1 );
        }   
        IF_FATAL_EXIT_S( i > TIMEOUT_SUNADC, "'CSigInputSunADC' (parant) timeout waiting for: ", PROG_SUNADC );			    
        // close write-end of pipe
        close( wrpipefd );
        FileDes = rdpipefd;
        StartedFlag = FALSE;
    }

}

// destructor: CSigInputSunADC
CSigInputSunADC::~CSigInputSunADC( )
{
    DEBUG_1_MSG("'CSigInputSunADC' derived destructor called" );
    
    if( FileDes >= 0 ) 
    {   // terminate child adc-proc. with SIGKILL
        kill( ChildPid, SIGKILL );
        close( FileDes );
    }
}

// start device SunADC
int CSigInputSunADC::startRead( const char *filename, int NrSkip )
{
    DEBUG_2_MSG("'CSigInputSunADC' startRead (ADC) called:");
    filename = 0;

    if( StartedFlag )
    {   WARNING_MSG("'CSigInputSunADC' startRead ADC: device already started" );
        return _ERR_;
    }

    // start adc proc. and first request with SIGUSR1
    End.setADC(FALSE);
    kill( ChildPid, SIGUSR1 );
    StartedFlag = TRUE;    
    SampleAnz = 0;

    return _OK_;
}

// stop device SunADC
int CSigInputSunADC::stopRead( EStopMode mode )
{
    DEBUG_2_MSG("'CSigInputSunADC' stopRead called" );

    // adc stop and clear
    if( mode == Stop )
    {   if( StartedFlag )
        {    // send SIGUSR2 to stop ADC, but keep samples in buffer for read out
            StartedFlag = False;
            kill( ChildPid, SIGUSR2 );   
        } else
        {    WARNING_MSG("'CSigInputSunADC' stopRead ADC: device not started" );
             return _ERR_;
        }
    }
    else if( StartedFlag ) 
    {   // send SIGINT to stop adc-proc. and clear samples in buffer 
        kill( ChildPid, SIGINT );
        short *dummy = new short[BlocksPerRequest*BlockSize];
        read( FileDes, dummy, BlocksPerRequest*BlockSize*sizeof(short) );
        StartedFlag = False;
    }

    return _OK_;
}

// read device SunAdc
int CSigInputSunADC::sigReadBlock( short *buf )
{
    DEBUG_4_MSG("'CSigInputSunADC' sigReadBlock called" );

    // read blocks from pipe
    char *chbuf = (char*)buf;
    long want = BlocksPerRequest*BlockSize*sizeof(short);
    long tries = 0, cnt  = 0, rd;
    while( cnt < want )
    {   rd = read( FileDes, chbuf+cnt, want - cnt );
        // if rd <= 0 child process may be terminated
        if( rd <= 0 )  
            tries++;
        else 
        {   tries = 0;
            cnt += rd;
        }
        if( tries > TRYREAD_SUNADC )            
        {    WARNING_MSG("'CSigInputSunADC' ADC process terminated during read" );
             return _ERR_;
        } 
    }
           
    SampleAnz += (cnt / sizeof(short));

    // test if child ended
    if( End.testADC() )
         return _ERR_;

    // else new request
    kill( ChildPid, SIGUSR1 );
    return _OK_;
}

//--------------------------- end of derived class 'CSigInputSunADC' ----------
#endif
//--------------------------- public derived class 'CSigInputAudioFile' -------

// constructor: CSigInputAudioFile
CSigInputAudioFile::CSigInputAudioFile( unsigned int blocksize )
:   CSigInput(blocksize, BL_PER_REQ_AUDIOFILE)
{
    DEBUG_1_MSG("'CSigInputAudioFile' derived constructor called" );
    Header.magic       = AUDIO_FILE_MAGIC;
    Header.hdr_size    = sizeof(Header);
    Header.data_size   = AUDIO_UNKNOWN_SIZE;
    Header.encoding    = ENCODING_AUDIOFILE;
    Header.sample_rate = 16000;
    Header.channels    = CHANNELS_AUDIOFILE;
}

// destructor: CSigInputAudioFile
CSigInputAudioFile::~CSigInputAudioFile( )
{
    DEBUG_1_MSG("'CSigInputAudioFile' derived destructor called" );
    
    if( FileDes >= 0 ) close( FileDes );
}

// open device AudioFile and read header
int CSigInputAudioFile::startRead( const char *filename, int NrSkip )
{
    if( filename == 0 )
        filename = "";
    DEBUG_2_MSG_S("'CSigInputAudioFile' startRead called:", filename );

    // close open stream
    if( FileDes >= 0 ) 
    {   close( FileDes );
        FileDes = -1;
        FrqSample = 0;
    }

    // open file (or stdin)
    if( *filename != '\0' )
    {   for( int i=0; i<3 && (FileDes = open( filename, O_RDONLY )) < 0; i++ ) ;
        IF_ERROR_RETURN_PERR_S( FileDes < 0, "'CSigInputAudioFile' startRead: can't open file;", filename );
    } else
        FileDes = 0;

    // read fixed part of header 
    if( read( FileDes, &Header, sizeof(Header) ) != sizeof(Header) )
    {   ERROR_PERR_S( "'CSigInputAudioFile' startRead: can't read audio-header from;", filename );
        close( FileDes );
        FileDes = -1;
        return _ERR_;
    }

    // test header
    IF_ERROR_RETURN_S( Header.magic != AUDIO_FILE_MAGIC, "'CSigInputAudioFile' startRead: this is not an audio-file;", filename );
    IF_ERROR_RETURN_S( Header.encoding != ENCODING_AUDIOFILE, "'CSigInputAudioFile' startRead: this audio-file is not 16-bit linear encoded;", filename );
    IF_ERROR_RETURN_S( Header.channels != CHANNELS_AUDIOFILE, "'CSigInputAudioFile' startRead: this audio-file is not one audio-channel;", filename );
	
    // seek over info-block	
    if( lseek (FileDes, Header.hdr_size+NrSkip, SEEK_SET ) < 0 )
    {   ERROR_PERR_S( "'CSigInputAudioFile' startRead: can't seek over audio-file-header;", filename );
        return _ERR_;
    }
    // now read-ptr is at begin of sample-data

    FrqSample = Header.sample_rate;
    SampleAnz = 0;

#if defined(DEBUG_LEVEL) && ( DEBUG_LEVEL>=3 || DEBUG_LEVEL==-3 )
 fprintf( stderr, ">>> audio-file: %s, HeaderSize=%d, DataSize=%d, FrqSamp=%d\n", filename, Header.hdr_size, Header.data_size, FrqSample );
#endif // DEBUG_LEVEL 3
    
    return _OK_;
}

// stop device, close audio-file
int CSigInputAudioFile::stopRead( EStopMode mode )
{
    DEBUG_2_MSG("'CSigInputAudioFile' stopRead called" );

    // close audio-file (stop and clear)
    if( mode == Stop && FileDes < 0)
    {   WARNING_MSG("'CSigInputAudioFile' stopRead: device already stoped" );
        return _ERR_;
    }
    else if( FileDes >= 0 ) 
    {  close(FileDes);
       FileDes = -1;
       FrqSample = 0;
    }

    return _OK_;
}

// read device audio-file
int CSigInputAudioFile::sigReadBlock( short *buf )
{
    DEBUG_4_MSG("'CSigInputAudioFile' sigReadBlock called" );

    // read block
    long rcount = read( FileDes, buf, BlocksPerRequest*BlockSize*sizeof(short) );
    long samples = rcount/sizeof(short);
    SampleAnz += samples;

    if( rcount !=  BlocksPerRequest*BlockSize*sizeof(short) )
        return _ERR_;
    else
        return _OK_;
}

// write block
int CSigInputAudioFile::writeBlock( int fd, short *buf )
{   
    // write block
    long wcount = write( fd, buf, BlocksPerRequest*BlockSize*sizeof(short) );
    if( wcount != BlocksPerRequest*BlockSize*sizeof(short) )
    {   ERROR_PERR( "'CSigInputAudioFile' writeBlock: write error" );
        return _ERR_;
    }

    return _OK_;
}

// write header
int CSigInputAudioFile::writeHeader( char *filename )
{   
    if( filename == 0 )
        filename = "";

    // open file (or stdout)
    int fd = 1;
    if( *filename != '\0' )
    {   fd = open( filename, O_WRONLY | O_CREAT );
        if( fd < 0 )
        {   ERROR_PERR_S("'CSigInputAudioFile' writeHeader: can't open file:", filename );
            return fd;
        }
    }

    // header write
    Header.hdr_size = sizeof(Header);
    if( write( fd, &Header, sizeof(Header) ) != sizeof(Header) )
    {   ERROR_PERR_S( "'CSigInputAudioFile' writeHeader: can't write header to:", filename );
        close( fd );
        return -1;
    }

    return fd;
}



//--------------------------- test --------------------------------------------

#if defined(TEST_LEVEL) && TEST_LEVEL == 1
#define BLOCKSIZE 160

int main( int argc, char *argv[] )
{   
    IF_ERROR_RETURN( argc != 2, "usage: siginput <file>" );  

    CSigInput *sigptr = new CSigInputFileILS( BLOCKSIZE );
    short *buf = new short[BLOCKSIZE*sigptr->getBlocksPerRequest()];

    sigptr->startRead( argv[1] );

    while( sigptr->sigReadBlock( buf ) == _OK_ );

    printf( "* %s: number of samples = %d\n", argv[1], sigptr->getSampleAnz() );
 
    delete [] buf;
    delete sigptr;

    return(0);
}


#endif  // TEST_LEVEL

#ifdef PRG_SWAPILS

#define BLOCKSIZE 256

int main( int argc, char *argv[] )
{   
    IF_ERROR_RETURN( argc != 3, "usage: swapils [<ils-file-in> <ils-file-out>]" );  

    CSigInputFileILS sig( BLOCKSIZE );
    short *buf = new short[BLOCKSIZE*sig.getBlocksPerRequest()];

    sig.startRead( argv[1] );
    int fd = sig.writeHeader( argv[2] );
    IF_FATAL_EXIT( fd<0, "output file open error" );

    while( sig.sigReadBlock( buf ) == _OK_ )
        if( sig.writeBlock( fd, buf) == _ERR_ )
            break;

    printf( "* %s: number of samples = %d\n", argv[1], sig.getSampleAnz() );
 
    delete [] buf;
    close(fd);
    return(0);
}


#endif  // PRG_SWAPILS

#if defined(TEST_LEVEL) && TEST_LEVEL == 2

#define BLOCKSIZE  160
static  CSigInputSunADC  sig( BLOCKSIZE );

class CSigInt : public Interrupt_handler
{
private:
    int   RunFlag;
    void  interrupt() { if(RunFlag) sig.stopRead(); else sig.startRead(NULL); RunFlag = !RunFlag; }
public:
    CSigInt() : Interrupt_handler(SIGINT), RunFlag(FALSE) {}
    int   isRun() const { return RunFlag; }
};


int main()
{
    // start tasten process here with fork (process dublication!)
    int childPid = fork();
    if( childPid == 0 )
    {
        // here: child process execution 
        char str[256];
        int  ppid = getppid();
        sleep(1);
        if( ppid == 1 ) exit(0);
        for(;;)
        {
        fprintf(stderr, "> siginput: press [return] to start recording\n" );
        if( gets(str) == NULL ) { kill(ppid,SIGKILL); exit(0); }
        kill(ppid, SIGINT);
        if( getppid() == 1 ) exit(0);
        fprintf(stderr, "> siginput: press [return] to stop recording\n" );
        if( gets(str) == NULL ) { kill(ppid,SIGKILL); exit(0); }
        kill(ppid, SIGINT);
        }
        exit(0);    
    } 

    long size = BLOCKSIZE*sig.getBlocksPerRequest();
    short *buf = new short[size];
    CSigInt  StSp;

    while( !StSp.isRun() ) pause();

    while( TRUE )
    {   if( sig.sigReadBlock( buf ) == _ERR_ )
            break;   
        if( write( 1, buf, size*sizeof(short) ) != size*sizeof(short) )
        {   fprintf( stderr, "> siginput: ERROR writing to stdout" );
            perror( "; " );
            kill(childPid,SIGKILL);
            exit(1);
        }
    }

    fprintf(stderr, "> siginput: recording %ld samples\n", sig.getSampleAnz() );

    delete [] buf;
    kill(childPid,SIGKILL);
    return 0;
}

#endif // TEST_LEVEL 2


#if defined(TEST_LEVEL) && TEST_LEVEL == 3

#define BLOCKSIZE 160

int main( int argc, char *argv[] )
{   
    IF_ERROR_RETURN( argc != 3, "usage: siginput <audio-file-in> <audio-file-out>" );  

    CSigInputAudioFile sig( BLOCKSIZE );
    short *buf = new short[BLOCKSIZE*sig.getBlocksPerRequest()];

    sig.startRead( argv[1] );
    int fd = sig.writeHeader( argv[2] );
    IF_FATAL_EXIT( fd<0, "output file open error" );

    while( sig.sigReadBlock( buf ) == _OK_ )
        if( sig.writeBlock( fd, buf) == _ERR_ )
            break;

    printf( "* %s: number of samples = %d\n", argv[1], sig.getSampleAnz() );
 
    delete [] buf;
    close(fd);
    return(0);
}


#endif  // TEST_LEVEL 3

#ifdef PRG_ILS_SUN

#define BLOCKSIZE 256

int main( int argc, char *argv[] )
{   
    if( argc > 3 )
    {   fprintf( stderr, "usage: ils_sun [<ils-file-in> <ils-file-out>]" );
        exit(-1);
    }

    char *fin  =  argc >= 2 ?  argv[1] : "" ;
    char *fout =  argc == 3 ?  argv[2] : "" ;

    CSigInputFileILS   sigin(  BLOCKSIZE );
    CSigInputAudioFile sigout( BLOCKSIZE );

    if( sigin.getBlocksPerRequest() != sigout.getBlocksPerRequest() )
    {   fprintf( stderr, "ERROR ils_sun: different blocksizes\n" );
        exit(-1);
    }

    long   bufsize = BLOCKSIZE*sigin.getBlocksPerRequest();
    short *buf = new short[bufsize];

    sigin.startRead( fin );
    int fd = sigout.writeHeader( fout );
    if( fd < 0 )
    {   perror( "ERROR ils_sun: output file open error; " );
        exit(-1);
    }

    while( sigin.sigReadBlock( buf ) == _OK_ )
        if( sigout.writeBlock( fd, buf) == _ERR_ )
            break;

 
    delete [] buf;
    close(fd);
    return(0);
}


#endif  // PRG_ILS_SUN

