/******************************************************************************
 * file:        specdoub.cc 
 * info:        implementation of CSpecDoubleBuf
 * written by:  M.Beham, 12.01.95 (org. from fft.c and sts.c )
 * last change: 
 * 
 ******************************************************************************/
/******************************************************************************
 *
 * Copyright (C) 1995 M. Beham
 *                                                                   
 * Alle Rechte vorbehalten. Diese Programmlistings und das daraus
 * erhaltene Programm duerfen nur mit schriftlicher Genehmigung	
 * des Autors weitergegeben werden.
 * Insbesondere duerfen auch Auszuege und die daraus erhaltenen
 * Programmteile nur mit schriftlicher Genehmigung des Autors
 * weitergeben, geaendert oder in Programmen anderer Autoren
 * verwendet werden.
 * Diese Copyright-Information darf nicht entfernt oder geaendert
 * werden.
 * 
 ******************************************************************************/
/******************************************************************************
 Beschreibung: Zwei reelle Signale mit einer FFT transformieren:

Die beiden Signalabschnitte "si1" und "si2" werden mit der Fensterfunktion
multipliziert und das Ergebnis in den Real-/ Imaginaerteil-Vektor
fuer die FFT geschrieben:
    si1 * window -> re  (Realteilvektor) 
    si2 * window -> im  (Imaginaerteilvektor)
 
Der complexe Vektor wird in einem Schritt mit dem FFT-Algorithmus 
Fourier-transformiert.

Zum Entflechten des comlexen FFT-Vektors in die beiden Teilspektren S1 vom 
Signalabschnitt si1 und S2 von si2 wird die gerade-Symmetrie des Spektrums
eines reellen Signals ausgenuetzt. S ist der komplexe FFT-Vektor beider 
Signale: si1 + j*si2 --> S. (siehe Murat Kunt, DSP, S.106f)
	Re{ S1(f) }  =  ( Re{ S(f) } + Re{ S(-f) } ) / 2
	Im{ S1(f) }  =  ( Im{ S(f) } - Im{ S(-f) } ) / 2
	Re{ S2(f) }  =  ( Im{ S(f) } + Im{ S(-f) } ) / 2
	Im{ S2(f) }  = -( Re{ S(f) } - Re{ S(-f) } ) / 2
  
Die Betragsquadratbildung und das Entflechten des komplexen Spektrums beider 
Signale erfolgt in einem Schritt (Breite der FFT: N,  i = 1,2,...N/2-1):
   4 * |S1(i)|^2  =  [ Re{S(i)} + Re{S(N-i)} ]^2 + [ Im{S(i)} - Im{S(N-i)} ]^2
   4 * |S2(i)|^2  =  [ Im{S(i)} + Im{S(N-i)} ]^2 + [ Re{S(i)} - Re{S(N-i)} ]^2

Der Faktor 1/4.0 kann bereits bei der Normierung der Fensterfunktion 
multiplikativ beruecksichtigt werden. Der Spektrale Anteil S1(i=0) und S2(i=0) 
(Gleichanteil) kann nicht aus dem verflochtenen komplexen Spektrum S(i) 
gewonnen werden und wird deshalb Null gesetzt:
	S1(0)  =  S2(0)  =  0

Die beiden Ergebnisse, Betrags-Quadrat FFT von si1 und si2, ueberschreiben
die uebergebenen Vektoren (RealVec und ImagVec) und haben nur noch eine Laenge 
von N/2.
 ******************************************************************************/




// #define DEBUG_LEVEL 4
// #define TEST_LEVEL 

#include "dbg_mess.h"
#include "specdoub.h"


//------------------------------ public ---------------------------------------

// constructor: CSpecDoubleBuf
CSpecDoubleBuf::CSpecDoubleBuf( unsigned int winlen, unsigned int fftlen )
 :  WinLen( winlen ),
    FFTLen( fftlen ),
    HamWin( winlen, 1.0/2.0 ),
    FFT( fftlen )
{
    DEBUG_1_MSG("'CSpecDoubleBuf' constructor called" );
    IF_FATAL_EXIT(  WinLen > FFTLen, "'CSpecDoubleBuf' FFT-length < window-length");
    // double buffer calloc 
    DoubleBuf[0] = new float[FFTLen];
    DoubleBuf[1] = new float[FFTLen];
    IF_FATAL_EXIT(  DoubleBuf[0] == 0 ||  DoubleBuf[1] == 0, "'CSpecDoubleBuf' out of memory");
}

// destructor: CSpecDoubleBuf
CSpecDoubleBuf::~CSpecDoubleBuf( )
{
    DEBUG_1_MSG("'CSpecDoubleBuf' destructor called" );

    delete [] DoubleBuf[0];
    delete [] DoubleBuf[1];
}

// calculate two power spec.
void  CSpecDoubleBuf::calcDoubleSpec()
{
    DEBUG_4_MSG("'CSpecDoubleBuf' calcDoubleSpec called" );

    // supress dc-level in both signal windows
    // HamWin.dcSub( DoubleBuf[0] );
    // HamWin.dcSub( DoubleBuf[1] );
    // already done in sigwin: Bit-Shift, conv. to float, Preemphase

    // count zero-cross-rate in both dc-free windows
    ZCRateBuf[0] = HamWin.countZC( DoubleBuf[0] );
    ZCRateBuf[1] = HamWin.countZC( DoubleBuf[1] );

    // multiply window function with both signals
    HamWin.mult( DoubleBuf[0] );
    HamWin.mult( DoubleBuf[1] );

    // fill rest = fftlen - winlen with 0.0
    if( FFTLen > WinLen )
        fillZero( );
    
    // fft transform, first calculate dc-values
    // float s0, s1;
    // calcDC( &s0, &s1 );
    FFT.transform( DoubleBuf[0], DoubleBuf[1] );

    // unfold spectrum and calculate power
    unfoldPowerSpec( );
    // DoubleBuf[0][0] = s0;
    // DoubleBuf[1][0] = s1;
}

//------------------------------ private CSpecDoubleBuf -----------------------

// fill rest of fft-length with zero
void  CSpecDoubleBuf::fillZero()
{
    register unsigned int i, n=FFTLen-WinLen;
    register float       *fp;

    for( fp = &DoubleBuf[0][WinLen], i = 0; i < n; i++ )
        *fp++ = (float)0.0;
    for( fp = &DoubleBuf[1][WinLen], i = 0; i < n; i++ )
        *fp++ = (float)0.0;
}
 
#define QUAD(x) ((x)*(x))

// unfold complex fft-spetrum of two real signals to two power spectra
void  CSpecDoubleBuf::unfoldPowerSpec()
{
    DEBUG_4_MSG("'CPowerSpecDouble' calcDouble called" );
    register float   *realptr = DoubleBuf[0];
    register float   *imagptr = DoubleBuf[1];
    register unsigned int i ,j;
    register float tmp;

    /* Entflechten der komplexen Spektren (siehe oben),
    *  Betragsquadrat bilden (Faktor 4.0 zu gross),
    *  das Ergebnis wird in die beiden uebergebenen Vektoren geschrieben (N/2),
    *  der Gleichanteil wird Null gesetzt */
    for( j=FFTLen-1, i=1; i < FFTLen/2; i++, j--)
    {
    	tmp        = QUAD(realptr[i]+realptr[j]) + QUAD(imagptr[i]-imagptr[j]);
    	imagptr[i] = QUAD(realptr[i]-realptr[j]) + QUAD(imagptr[i]+imagptr[j]);
        realptr[i] = tmp;
    }
    realptr[0] = imagptr[0] = 0.0;
}

// calculate DC-Values
void  CSpecDoubleBuf::calcDC( float *s0, float *s1 )
{
    register int i;
    register float *p0 = DoubleBuf[0], *p1 = DoubleBuf[1];

    for( i=0; i<FFTLen; i++ )
    {   *s0 += *p0++;
        *s1 += *p1++; 
    }
}
//------------------------------ End of CSpecDoubleBuf ------------------------

//------------------------------ Test -----------------------------------------

#ifdef TEST_LEVEL

#include <math.h>
#include <stdio.h>

int main()
{   unsigned int winlen, fftlen;

    printf( "\nWin length: " ); scanf( "%d", &winlen );
    printf( "\nFFT length: " ); scanf( "%d", &fftlen );

    CSpecDoubleBuf buf(winlen, fftlen);
    float f0,f1,pim = (2.0*M_PI)/winlen;
   
    printf("\nSignal 0, amp = 16.0, sin f0 k*[fs/N]: "); scanf( "%f", &f0 );
    printf("\nSignal 1, amp = 32.0, cos f1 k*[fs/N]: "); scanf( "%f", &f1 );

    for( int i = 0; i < winlen; i++ ) 
    {   buf.getBuffer(0)[i] = 16.0 * sin( pim*f0*i );
        buf.getBuffer(1)[i] = 32.0 * cos( pim*f1*i );
    }

    buf.calcDoubleSpec();

    printf("\nlog. spec0:\n");
    for( i = 0; i < winlen/2; i++ ) printf( "%5.1f ", buf.getBuffer(0)[i] );
    printf("\nlog. spec1:\n");
    for( i = 0; i < winlen/2; i++ ) printf( "%5.1f ", buf.getBuffer(1)[i] );

    return 0;
}

#endif  // TEST_LEVEL
